"""This module creates running statistics for timerseries"""

#pylint: disable=line-too-long
#pylint: disable=trailing-whitespace
#pylint: disable=invalid-name

import pandas as pd
import numpy as np
import matplotlib
import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle
import seaborn as sn
import cmasher as cmr

from SuPerSim.pickling import load_all_pickles
from SuPerSim.constants import colorcycle
from SuPerSim.open import open_thaw_depth_nc

def table_background_evolution_mean_GST_aspect_slope(site, path_pickle, path_thaw_depth=None):
    """ Function returns dataframes of mean background and evolution of GST (ground-surface temperature)
        as a function of slope, aspect, and altitude
        same for number of simulations per cell and permafrost state
        and a dictionary of values of ['aspect', 'slope','altitude', 'forcing']

    Parameters
    ----------
    site : str
        Location of the event, e.g. 'Aksaut_North', will be used to label all the pickles
    path_pickle : str
        String path to the location of the folder where the pickles are saved
    path_thaw_depth : str
        Path to the .nc file where the aggregated thaw depth simulations are stored
    
    Returns
    -------
    table_all : dict
        Dictionray containing:
            df_bkg : dict of pandas.core.frame.DataFrame
                Dictionary assigning a dataframe of background GST for each aspect and slope
                e.g. df_bkg = {2900: panda with columns=aspect and rows=slope, 3000: ...}
            df_evol : dict of pandas.core.frame.DataFrame
                Average GST evolution between background and transient over all simulations in that cell
            df_num_sim : dict of pandas.core.frame.DataFrame
                Number of simulations in that cell
            df_perma : dict of pandas.core.frame.DataFrame
                Permafrost state in that cell, in ['glacier', 'permafrost', 'no permafrost']
            dic_var : dict
                Dictionary of values of ['aspect', 'slope','altitude', 'forcing'], e.g
                dic_var = {'aspect': array([22.5, 45. , 67.5]), 'slope': array([55, 60, 65, 70, 75]), ...}
    """

    pkl = load_all_pickles(site, path_pickle)
    df_stats = pkl['df_stats']
    list_valid_sim = pkl['list_valid_sim']


    variables = ['aspect', 'slope','altitude', 'forcing']
    dic_var = {}
    dic_var = {i: np.sort(np.unique(df_stats.loc[:, i], return_counts=False)) for i in variables}

    if path_thaw_depth is not None:
        _, thaw_depth = open_thaw_depth_nc(path_thaw_depth)
        list_valid_sim = list(df_stats.index.values)
        list_no_perma = []
        for sim in list_valid_sim:
            if np.std(thaw_depth[sim,:]) < 1 and np.max(thaw_depth[sim,:])> 19:
                list_no_perma.append(sim)
        list_perma = list(set(list_valid_sim) - set(list_no_perma))

    list_grd_temp = {altitude: {slope: {aspect: [] for aspect in dic_var['aspect']} for slope in dic_var['slope']} for altitude in dic_var['altitude']}
    list_mean_grd_temp = {altitude: {slope: {aspect: [] for aspect in dic_var['aspect']} for slope in dic_var['slope']} for altitude in dic_var['altitude']}
    list_diff_temp = {altitude: {slope: {aspect: [] for aspect in dic_var['aspect']} for slope in dic_var['slope']} for altitude in dic_var['altitude']}
    list_mean_diff_temp = {altitude: {slope: {aspect: [] for aspect in dic_var['aspect']} for slope in dic_var['slope']} for altitude in dic_var['altitude']}
    list_num_sim = {altitude: {slope: {aspect: [] for aspect in dic_var['aspect']} for slope in dic_var['slope']} for altitude in dic_var['altitude']}
    check_perma_list = {altitude: {slope: {aspect: [] for aspect in dic_var['aspect']} for slope in dic_var['slope']} for altitude in dic_var['altitude']}

    for altitude in dic_var['altitude']:
        for slope in dic_var['slope']:
            for aspect in dic_var['aspect']:
                df_this_sim = df_stats[(df_stats['aspect']==aspect) & (df_stats['slope']==slope) & (df_stats['altitude']==altitude)]
                list_grd_temp[altitude][slope][aspect] = list(df_this_sim['bkg_grd_temp'])
                list_sim_per_forcing = [list(df_this_sim['forcing']).count(i) for i in dic_var['forcing']]
                list_num_sim[altitude][slope][aspect] = (np.sum(list_sim_per_forcing) if len(set(list_sim_per_forcing))==1 else np.nan)
                list_mean_grd_temp[altitude][slope][aspect] = round(np.mean((list_grd_temp[altitude][slope][aspect])),3)
                list_diff_temp[altitude][slope][aspect] = list(df_this_sim['trans_grd_temp'] - df_this_sim['bkg_grd_temp'])
                list_mean_diff_temp[altitude][slope][aspect] = round(np.mean((list_diff_temp[altitude][slope][aspect])),3)
                if path_thaw_depth is not None:
                    temp_list = [i in list_perma for i in list(df_stats[(df_stats['altitude']==altitude) & (df_stats['aspect']==aspect) & (df_stats['slope']==slope)].index.values)]
                    check_perma_list[altitude][slope][aspect] = ('no permafrost' if len(temp_list)<3 else ('glacier' if all(temp_list) else 'permafrost'))
                else:
                    check_perma_list[altitude][slope][aspect] = np.nan

    df_bkg = {altitude: pd.DataFrame(list_mean_grd_temp[altitude]).T for altitude in dic_var['altitude']}
    df_evol = {altitude: pd.DataFrame(list_mean_diff_temp[altitude]).T for altitude in dic_var['altitude']}
    df_num_sim = {altitude: pd.DataFrame(list_num_sim[altitude]).T for altitude in dic_var['altitude']}
    df_perma = {altitude: pd.DataFrame(check_perma_list[altitude]).T for altitude in dic_var['altitude']}

    table_all = {'bkg': df_bkg, 'evol': df_evol, 'num_sim': df_num_sim, 'perma': df_perma, 'vars': dic_var}

    return table_all

def prepare_data_table_GST(site, path_pickle, path_thaw_depth=None):
    """ Function returns all data needed to produce topographic heatmaps
    
    Parameters
    ----------
    site : str
        Location of the event, e.g. 'Joffre' or 'Fingerpost'
    path_pickle : str
        String path to the location of the folder where the pickles are saved
    path_thaw_depth : str
        Path to the .nc file where the aggregated thaw depth simulations are stored


    Returns
    -------
    table_all : dict
        Dictionray containing:
            df_bkg : dict of pandas.core.frame.DataFrame
                Dictionary assigning a dataframe of background GST for each aspect and slope
                e.g. df_bkg = {2900: panda with columns=aspect and rows=slope, 3000: ...}
            df_evol : dict of pandas.core.frame.DataFrame
                Average GST evolution between background and transient over all simulations in that cell
            df_num_sim : dict of pandas.core.frame.DataFrame
                Number of simulations in that cell
            df_perma : dict of pandas.core.frame.DataFrame
                Permafrost state in that cell, in ['glacier', 'permafrost', 'no permafrost']
            dic_var : dict
                Dictionary of values of ['aspect', 'slope','altitude', 'forcing'], e.g
                dic_var = {'aspect': array([22.5, 45. , 67.5]), 'slope': array([55, 60, 65, 70, 75]), ...}
    rockfall_values : dict
        Dictionary with date and topography info for the event
    sim_per_cell : int
        number of simulations for a given cell with fixed (altitude, slope, aspect) triplet
    """

    pkl = load_all_pickles(site, path_pickle)
    df = pkl['df']
    rockfall_values = pkl['rockfall_values']

    table_all = table_background_evolution_mean_GST_aspect_slope(site, path_pickle, path_thaw_depth)
    dic_var = table_all['vars']
    
    variables = ['aspect', 'slope', 'altitude']
    # number of simulations for a given cell with fixed (altitude, slope, aspect) triplet
    sim_per_cell = int(len(df)/np.prod([len(dic_var[i]) for i in variables]))

    return table_all, rockfall_values, sim_per_cell

def plot_table_mean_GST_aspect_slope_single_altitude(table_all, rockfall_values, altitude, background=True, box=True):
    """ Function returns a plot of the table of either mean background GST (ground-surface temperature)
        or its evolution between the background and the transient periods,
        as a function of slope and aspect, for a given altitude and higlights the cell corresponding to the 
        rockfall starting zone
    
    Parameters
    ----------
    table_all : dict
        Dictionray containing:
            df_bkg : dict of pandas.core.frame.DataFrame
                Dictionary assigning a dataframe of background GST for each aspect and slope
                e.g. df_bkg = {2900: panda with columns=aspect and rows=slope, 3000: ...}
            df_evol : dict of pandas.core.frame.DataFrame
                Average GST evolution between background and transient over all simulations in that cell
            df_num_sim : dict of pandas.core.frame.DataFrame
                Number of simulations in that cell
            df_perma : dict of pandas.core.frame.DataFrame
                Permafrost state in that cell, in ['glacier', 'permafrost', 'no permafrost']
            dic_var : dict
                Dictionary of values of ['aspect', 'slope','altitude', 'forcing'], e.g
                dic_var = {'aspect': array([22.5, 45. , 67.5]), 'slope': array([55, 60, 65, 70, 75]), ...}
    rockfall_values : dict
        Dictionary with date and topography info for the event
    altitude :
        desired altitude for the plot
    background : bool, optional 
        If True, plots the mean background value, else, plots the evolution of the mean
    box : bool, optional 
        If True, highlights the cell corresponding to the rockfall starting zone with a black box


    Returns
    -------
    fig : Figure
        heatmap of background or evolution of GST
    """

    list_mean = (table_all['bkg'][altitude] if background else table_all['evol'][altitude])

    vals = np.around(list_mean.values, 2)
    vals = vals[~np.isnan(vals)]
    dilute = 1
    min_vals = np.min(vals)
    max_vals = np.max(vals)
    range_vals = max_vals- min_vals
    normal = plt.Normalize((np.max(max_vals - 2*dilute*range_vals,0) if np.min(vals)>0 else -dilute*np.max(np.abs(vals))), dilute*np.max(np.abs(vals)))
    colours = plt.cm.seismic(normal(list_mean)) #pylint: disable=no-member

    for aspect in list_mean:
        for slope in list_mean.index:
            if np.isnan(list_mean[aspect].loc[slope]):
                colours[aspect][slope] = list(matplotlib.colors.to_rgba('silver'))

    fig = plt.figure()
    ax = fig.add_subplot(111, frameon=False, xticks=[], yticks=[])

    the_table=plt.table(cellText=[list(list_mean.loc[i]) for i in list_mean.index], rowLabels=list_mean.index, colLabels=list_mean.columns, 
                        loc='center', cellLoc='center',
                        cellColours=colours)

    if box and rockfall_values['exact_topo']:
        if altitude == rockfall_values['altitude']:
            ax.add_patch(Rectangle(((rockfall_values['aspect'])/45*1/8, (70-rockfall_values['slope'])/10*1/6), 1/8, 1/6,
                        edgecolor = 'black', transform=ax.transAxes,
                        fill=False,
                        lw=4))

    the_table.scale(1, 3.7)
    the_table.set_fontsize(16)
    ax.axis('off')

    plt.text(-0.08, 5/12,'Slope [°]', fontsize= 16, rotation=90, horizontalalignment='right', verticalalignment='center')
    plt.text(0.5, 1,'Aspect [°]', fontsize= 16, rotation=0, horizontalalignment='center', verticalalignment='bottom')

    plt.tight_layout()
    plt.show()
    plt.close()

    return fig

def plot_table_mean_GST_aspect_slope_single_altitude_from_inputs(site, path_pickle, altitude, background=True, box=True):
    """ Function returns a plot of the table of either mean background GST (ground-surface temperature)
        or its evolution between the background and the transient periods,
        as a function of slope and aspect, for a given altitude and higlights the cell corresponding to the 
        rockfall starting zone
    
    Parameters
    ----------
    site : str
        Location of the event, e.g. 'Joffre' or 'Fingerpost'
    path_pickle : str
        String path to the location of the folder where the pickles are saved
    altitude :
        desired altitude for the plot
    background : bool, optional 
        If True, plots the mean background value, else, plots the evolution of the mean
    box : bool, optional 
        If True, highlights the cell corresponding to the rockfall starting zone with a black box

    Returns
    -------
    fig : Figure
        heatmap of background or evolution of GST
    """

    table_all, rockfall_values, _ = prepare_data_table_GST(site, path_pickle)
    fig = plot_table_mean_GST_aspect_slope_single_altitude(table_all, rockfall_values, altitude, background, box)

    return fig

def plot_table_mean_GST_aspect_slope_all_altitudes(table_all, rockfall_values, sim_per_cell, show_glaciers=True, box=True):
    """ Function returns a plot of the table of either mean background GST (ground-surface temperature)
        or its evolution between the background and the transient periods,
        as a function of slope, aspect, and altitude and higlights the cell corresponding to the 
        rockfall starting zone
    
    Parameters
    ----------
    table_all : dict
        Dictionray containing:
            df_bkg : dict of pandas.core.frame.DataFrame
                Dictionary assigning a dataframe of background GST for each aspect and slope
                e.g. df_bkg = {2900: panda with columns=aspect and rows=slope, 3000: ...}
            df_evol : dict of pandas.core.frame.DataFrame
                Average GST evolution between background and transient over all simulations in that cell
            df_num_sim : dict of pandas.core.frame.DataFrame
                Number of simulations in that cell
            df_perma : dict of pandas.core.frame.DataFrame
                Permafrost state in that cell, in ['glacier', 'permafrost', 'no permafrost']
            dic_var : dict
                Dictionary of values of ['aspect', 'slope','altitude', 'forcing'], e.g
                dic_var = {'aspect': array([22.5, 45. , 67.5]), 'slope': array([55, 60, 65, 70, 75]), ...}
    rockfall_values : dict
        Dictionary with date and topography info for the event
    sim_per_cell : int
        number of simulations for a given cell with fixed (altitude, slope, aspect) triplet
    show_glaciers : bool, opional
        Whether or not to plot the glacier fraction
    box : bool, optional 
        If True, highlights the cell corresponding to the rockfall starting zone with a black box


    Returns
    -------
    fig : Figure
        (2 or 3)*(# altitudes) heatmaps of background GST, evolution of GST, and possibly number of glaciers
    """

    dic_var = table_all['vars']

    alt_list = list(dic_var['altitude'])
    for i,j in enumerate(alt_list):
        if box and rockfall_values['exact_topo']:
            if rockfall_values['altitude'] == j:
                alt_index = i

    # setting the parameter values 
    annot = True
    center = [0, 0]
    cmap = ['seismic', 'seismic']

    list_mean = [# evolution mean GST
                table_all['evol'],
                # background mean GST
                table_all['bkg']]

    labels_plot = ['Mean GST evolution [°C]', 'Mean background GST [°C]']
    if show_glaciers:
        # glacier fraction
        list_mean.append({alt: table_all['num_sim'][alt].transform(lambda x: ((sim_per_cell-x)/sim_per_cell*100)).astype(int) for alt in alt_list})
        labels_plot.append('Glacier fraction')
        cmap.append('BrBG')
        center.append(50)

    nrows= len(list_mean)
    ncols = len(list_mean[0])

    fig, axs = plt.subplots(nrows=nrows, ncols=ncols, figsize=(5*ncols, 3*nrows), constrained_layout=True)
    no_nan = [[i[altitude][aspect][slope]
            for altitude in dic_var['altitude'] for aspect in dic_var['aspect'] for slope in dic_var['slope']
            if not np.isnan(i[altitude][aspect][slope])]
            for i in list_mean]
    vmin = [np.min(i) for i in no_nan]
    vmax = [np.max(i) for i in no_nan]
    if len(vmin) == 3:
        vmin[2] = 0
        vmax[2] = 100

    # plotting the heatmap 
    for j in range(nrows):
        if ncols == 1:
            sn.heatmap(data=list_mean[nrows-1-j][alt_list[0]], annot=annot, center=center[nrows-1-j], cmap=cmap[nrows-1-j], ax=axs[j], vmin=vmin[nrows-1-j], vmax=vmax[nrows-1-j],
                    cbar=True, yticklabels=True, xticklabels=(j==nrows-1), cbar_kws={'label': labels_plot[nrows-1-j]})
            axs[0].figure.axes[-1].yaxis.label.set_size(13)
            if box and rockfall_values['exact_topo']:
                axs[j].add_patch(Rectangle(((rockfall_values['aspect'])/45*1/8, (70-rockfall_values['slope'])/10*1/5), 1/8, 1/5,
                                        edgecolor = 'black', transform=axs[j].transAxes, fill=False, lw=4))
        if ncols > 1:
            for i in range(ncols):
                sn.heatmap(data=list_mean[nrows-1-j][alt_list[i]], annot=annot, center=center[nrows-1-j], cmap=cmap[nrows-1-j], ax=axs[j,i], vmin=vmin[nrows-1-j], vmax=vmax[nrows-1-j],
                            cbar=(i==ncols-1), yticklabels=(i==0), xticklabels=(j==nrows-1), cbar_kws={'label': labels_plot[nrows-1-j]})
            axs[0,0].figure.axes[-1].yaxis.label.set_size(13)
            if box and rockfall_values['exact_topo']:
                axs[j,alt_index].add_patch(Rectangle(((rockfall_values['aspect'])/45*1/8, (70-rockfall_values['slope'])/10*1/5), 1/8, 1/5,
                                                    edgecolor = 'black', transform=axs[j,alt_index].transAxes, fill=False, lw=4))

    if ncols == 1:
        axs[0].set_title(f'{alt_list[0]} m')
    if ncols > 1:
        for i in range(ncols):
            axs[0,i].set_title(f'{alt_list[i]} m')
    fig.supxlabel('Aspect [°]')
    fig.supylabel('Slope [°]')

    # displaying the plotted heatmap 
    plt.show()
    plt.close()

    return fig

def plot_table_mean_GST_aspect_slope_all_altitudes_from_inputs(site, path_pickle, show_glaciers=True, box=True):
    """ Function returns a plot of the table of either mean background GST (ground-surface temperature)
        or its evolution between the background and the transient periods,
        as a function of slope, aspect, and altitude and higlights the cell corresponding to the 
        rockfall starting zone
    
    Parameters
    ----------
    site : str
        Location of the event, e.g. 'Joffre' or 'Fingerpost'
    path_pickle : str
        String path to the location of the folder where the pickles are saved
    altitude :
        desired altitude for the plot
    show_glaciers : bool, opional
        Whether or not to plot the glacier fraction
    box : bool, optional 
        If True, highlights the cell corresponding to the rockfall starting zone with a black box

    Returns
    -------
    fig : Figure
        (2 or 3)*(# altitudes) heatmaps of background GST, evolution of GST, and possibly number of glaciers
    """

    table_all, rockfall_values, sim_per_cell = prepare_data_table_GST(site, path_pickle)
    fig = plot_table_mean_GST_aspect_slope_all_altitudes(table_all, rockfall_values, sim_per_cell, show_glaciers, box)

    return fig

def plot_table_mean_GST_aspect_slope_all_altitudes_polar(table_all, rockfall_values, box=True):
    """ Function returns a polar plot per altitude of the table of either mean background GST (ground-surface temperature)
        or its evolution between the background and the transient periods,
        as a function of slope, aspect, and altitude and higlight the cell corresponding to the 
        rockfall starting zone
    
    Parameters
    ----------
    table_all : dict
        Dictionray containing:
            df_bkg : dict of pandas.core.frame.DataFrame
                Dictionary assigning a dataframe of background GST for each aspect and slope
                e.g. df_bkg = {2900: panda with columns=aspect and rows=slope, 3000: ...}
            df_evol : dict of pandas.core.frame.DataFrame
                Average GST evolution between background and transient over all simulations in that cell
            df_num_sim : dict of pandas.core.frame.DataFrame
                Number of simulations in that cell
            df_perma : dict of pandas.core.frame.DataFrame
                Permafrost state in that cell, in ['glacier', 'permafrost', 'no permafrost']
            dic_var : dict
                Dictionary of values of ['aspect', 'slope','altitude', 'forcing'], e.g
                dic_var = {'aspect': array([22.5, 45. , 67.5]), 'slope': array([55, 60, 65, 70, 75]), ...}
    rockfall_values : dict
        Dictionary with date and topography info for the event
    box : bool, optional 
        If True, highlights the cell corresponding to the rockfall starting zone with a black box

    Returns
    -------
    fig : Figure
        2*(# altitudes) polar heatmaps of background GST and evolution of GST
    """

    dic_var = table_all['vars']

    alt_list = dic_var['altitude']
    for i,j in enumerate(alt_list):
        if box and rockfall_values['exact_topo']:
            if rockfall_values['altitude'] == j:
                alt_index = i

    list_mean = [table_all['bkg'], table_all['evol']]
    no_nan = [[i[altitude][aspect][slope]
                for altitude in dic_var['altitude'] for aspect in dic_var['aspect'] for slope in dic_var['slope']
                if not np.isnan(i[altitude][aspect][slope])]
                for i in list_mean]
    vmin = [np.min(i) for i in no_nan]
    vmax = [np.max(i) for i in no_nan]

    list_slope = dic_var['slope']
    list_aspect = dic_var['aspect']

    nbin_slope = len(list_slope)
    delta_slope = list_slope[1] - list_slope[0]
    nbin_aspect = len(list_aspect) # int(36/delta_aspect)
    delta_aspect = list_aspect[1] - list_aspect[0]

    # binning
    rbins = np.linspace(np.min(list_slope), np.max(list_slope), nbin_slope)
    abins = np.linspace(list_aspect[0]/360*2*np.pi, list_aspect[-1]/360*2*np.pi, nbin_aspect)
    subdivs = 100
    abins2 = np.linspace((list_aspect[0]-delta_aspect/2)/360*2*np.pi, (list_aspect[-1]+delta_aspect/2)/360*2*np.pi, subdivs*nbin_aspect)
    A, R = np.meshgrid(abins2, rbins)

    nrows = len(list_mean)
    ncols = len(list_mean[0])

    fig, axs = plt.subplots(ncols=ncols, nrows=nrows, figsize=(5*ncols, 5*nrows), subplot_kw=dict(projection="polar"),
                            gridspec_kw={'hspace': -0.1, 'wspace': 0.3})
    abs_max = [np.max(np.abs([vmin[i], vmax[i]])) for i in range(len(vmin))]

    cmap = [cmr.get_sub_cmap('seismic', 0.5 + 0.5*(vmin[i]/abs_max[i]), 0.5 + 0.5*(vmax[i]/abs_max[i])) for i in range(len(vmin))]

    ticks = [[],[]]
    tick_pos = [[],[]]
    for i,v in enumerate(vmin):
        for space in [0.05, 0.1, 0.2, 0.5, 1]:
            pre_ticks = np.arange(np.ceil(v/space)*space+0, np.min([vmax[i], np.floor(vmax[i]/space+1)*space]), space)
            if (len(pre_ticks)>=5 and len(pre_ticks)<10):
                ticks[i] = [round(j,(2 if space<0.1 else 1)) for j in pre_ticks]
        tick_pos[i] = [(j-v)/(vmax[i]-v) for j in ticks[i] if (j>=v and j<vmax[i])]

    for j in range(nrows):
        for i,alt in enumerate(dic_var['altitude']):
            temp_list = [list(list_mean[j][alt].loc[k]) for k in list_mean[j][alt].index]
            axs[j,i].pcolormesh(A, R, np.repeat(temp_list, subdivs, axis=1), cmap=cmap[j], vmin=vmin[j], vmax=vmax[j])
            axs[j,i].set_facecolor("silver")
            axs[j,i].set_theta_zero_location('N')
            axs[j,i].set_theta_direction(-1)
            axs[j,i].tick_params(axis='y', labelcolor='black')
            axs[j,i].set_xticks(abins)
            axs[j,i].set_yticks([])
            axs[j,i].xaxis.grid(False)
            axs[j,i].yaxis.grid(False)
            axs[j,i].set_rlim(int(np.max(list_slope)+delta_slope),int(np.min(list_slope)-delta_slope))
            axs[0,i].set_title(f'{alt_list[i]} m')
            axs[j,i].bar(0, 1).remove()
            for k in range(int(np.min(list_slope)), int(np.max(list_slope))+1,
                        int((np.max(list_slope)-np.min(dic_var['slope']))/(nbin_slope-1))):
                axs[j,i].text(np.pi/5, k, (f'{k}°'), horizontalalignment='center', verticalalignment='center')
        if box and rockfall_values['exact_topo']:
            axs[j,alt_index].add_patch(matplotlib.patches.Rectangle(((rockfall_values['aspect']-45/2)/360*2*np.pi, rockfall_values['slope']-5),
                                                        width=np.pi/4, height=10, edgecolor = 'black', fill=False, lw=2))
        

    for j in range(nrows):
        #pylint: disable=no-member
        fig.colorbar(plt.cm.ScalarMappable(cmap=cmap[j]), shrink=.7,
                        ax=axs[j,:].ravel().tolist(), orientation='vertical',
                        label=('Mean background GST [°C]' if j==0 else 'Mean GST evolution [°C]'))

    for idx,i in enumerate(range(-nrows,0)):
        axs[0,0].figure.axes[i].yaxis.set_ticks(tick_pos[idx])
        axs[0,0].figure.axes[i].set_yticklabels(ticks[idx]) 

    for i in [6,7]:
        axs[0,0].figure.axes[i].yaxis.label.set_size(13)

    plt.show()
    plt.close()

    return fig

def plot_table_mean_GST_aspect_slope_all_altitudes_polar_from_inputs(site, path_pickle, box=True):
    """ Function returns a polar plot per altitude of the table of either mean background GST (ground-surface temperature)
        or its evolution between the background and the transient periods,
        as a function of slope, aspect, and altitude and higlight the cell corresponding to the 
        rockfall starting zone
    
    Parameters
    ----------
    site : str
        Location of the event, e.g. 'Joffre' or 'Fingerpost'
    path_pickle : str
        String path to the location of the folder where the pickles are saved
    box : bool, optional 
        If True, highlights the cell corresponding to the rockfall starting zone with a black box

    Returns
    -------
    fig : Figure
        2*(# altitudes) polar heatmaps of background GST and evolution of GST
    """

    table_all, rockfall_values, _ = prepare_data_table_GST(site, path_pickle)
    fig = plot_table_mean_GST_aspect_slope_all_altitudes_polar(table_all, rockfall_values, box)

    return fig

def plot_permafrost_all_altitudes_polar(table_all, rockfall_values, box=True):
    """ Function returns a polar plot per altitude of the table of the permafrost state
        as a function of slope, aspect, and altitude and higlight the cell corresponding to the 
        rockfall starting zone
    
    Parameters
    ----------
    table_all : dict
        Dictionray containing:
            df_bkg : dict of pandas.core.frame.DataFrame
                Dictionary assigning a dataframe of background GST for each aspect and slope
                e.g. df_bkg = {2900: panda with columns=aspect and rows=slope, 3000: ...}
            df_evol : dict of pandas.core.frame.DataFrame
                Average GST evolution between background and transient over all simulations in that cell
            df_num_sim : dict of pandas.core.frame.DataFrame
                Number of simulations in that cell
            df_perma : dict of pandas.core.frame.DataFrame
                Permafrost state in that cell, in ['glacier', 'permafrost', 'no permafrost']
            dic_var : dict
                Dictionary of values of ['aspect', 'slope','altitude', 'forcing'], e.g
                dic_var = {'aspect': array([22.5, 45. , 67.5]), 'slope': array([55, 60, 65, 70, 75]), ...}
    rockfall_values : dict
        Dictionary with date and topography info for the event
    box : bool, optional 
        If True, highlights the cell corresponding to the rockfall starting zone with a black box

    Returns
    -------
    fig : Figure
        (# altitudes) polar heatmaps of permafrost state
    """

    dic_var = table_all['vars']

    alt_list = dic_var['altitude']
    for i,j in enumerate(alt_list):
        if box and rockfall_values['exact_topo']:
            if rockfall_values['altitude'] == j:
                alt_index = i

    data = table_all['perma']
    data_to_num = {alt: data[alt].replace(['glacier', 'permafrost', 'no permafrost'], [1/6, 1/2, 5/6]) for alt in data.keys()}

    list_slope = dic_var['slope']
    list_aspect = dic_var['aspect']

    nbin_slope = len(list_slope)
    delta_slope = list_slope[1] - list_slope[0]
    nbin_aspect = len(list_aspect) # int(36/delta_aspect)
    delta_aspect = list_aspect[1] - list_aspect[0]

    # binning
    rbins = np.linspace(np.min(list_slope), np.max(list_slope), nbin_slope)
    abins = np.linspace(list_aspect[0]/360*2*np.pi, list_aspect[-1]/360*2*np.pi, nbin_aspect)
    subdivs = 100
    abins2 = np.linspace((list_aspect[0]-delta_aspect/2)/360*2*np.pi, (list_aspect[-1]+delta_aspect/2)/360*2*np.pi, subdivs*nbin_aspect)
    A, R = np.meshgrid(abins2, rbins)

    ncols = len(data)
    nrows = 1

    fig, axs = plt.subplots(ncols=ncols, nrows=nrows, figsize=(5*ncols, 5*nrows), subplot_kw=dict(projection="polar"),
                            gridspec_kw={'hspace': -0.1, 'wspace': 0.3})

    cmap = matplotlib.colors.ListedColormap([colorcycle[1],colorcycle[2],colorcycle[0]]) 

    for i,alt in enumerate(dic_var['altitude']):
        axs[i].pcolormesh(A, R, np.repeat(data_to_num[alt], subdivs, axis=1), cmap=cmap, vmin=0, vmax=1)
        axs[i].set_facecolor("silver")
        axs[i].set_theta_zero_location('N')
        axs[i].set_theta_direction(-1)
        axs[i].tick_params(axis='y', labelcolor='black')
        axs[i].set_xticks(abins)
        axs[i].set_yticks([])
        axs[i].xaxis.grid(False)
        axs[i].yaxis.grid(False)
        axs[i].set_rlim(int(np.max(list_slope)+delta_slope),int(np.min(list_slope)-delta_slope))
        axs[i].set_title(f'{alt_list[i]} m')
        axs[i].bar(0, 1).remove()
        for k in range(int(np.min(list_slope)), int(np.max(list_slope))+1,
                        int((np.max(list_slope)-np.min(dic_var['slope']))/(nbin_slope-1))):
            axs[i].text(np.pi/5, k, (f'{k}°'), horizontalalignment='center', verticalalignment='center')
    if box and rockfall_values['exact_topo']:
        axs[alt_index].add_patch(matplotlib.patches.Rectangle(((rockfall_values['aspect']-45/2)/360*2*np.pi, rockfall_values['slope']-5),
                                                        width=np.pi/4, height=10, edgecolor = 'black', fill=False, lw=2))

    #pylint: disable=no-member
    fig.colorbar(plt.cm.ScalarMappable(cmap=cmap),
                    shrink=.8, ax=axs[:].ravel().tolist(), orientation='vertical')
    axs[0].figure.axes[-1].yaxis.set_ticks([1/6,1/2,5/6])
    axs[0].figure.axes[-1].set_yticklabels(['Permafrost', 'No permafrost, no glaciers', 'Glaciers']) 
    axs[0].figure.axes[-1].yaxis.label.set_size(15)

    plt.show()
    plt.close()

    return fig

def plot_permafrost_all_altitudes_polar_from_inputs(site, path_pickle, path_thaw_depth, box=True):
    """ Function returns a polar plot per altitude of the table of the permafrost state
        as a function of slope, aspect, and altitude and higlight the cell corresponding to the 
        rockfall starting zone
    
    Parameters
    ----------
    site : str
        Location of the event, e.g. 'Joffre' or 'Fingerpost'
    path_pickle : str
        String path to the location of the folder where the pickles are saved
    path_thaw_depth : str
        Path to the .nc file where the aggregated thaw depth simulations are stored
    box : bool, optional 
        If True, highlights the cell corresponding to the rockfall starting zone with a black box

    Returns
    -------
    fig : Figure
        (# altitudes) polar heatmaps of permafrost state
    """
    table_all, rockfall_values, _ = prepare_data_table_GST(site, path_pickle, path_thaw_depth)
    fig = plot_permafrost_all_altitudes_polar(table_all, rockfall_values, box)

    return fig



"""This module creates weights for averaging timeseries"""

#pylint: disable=line-too-long
#pylint: disable=trailing-whitespace
#pylint: disable=invalid-name

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

from SuPerSim.open import open_thaw_depth_nc
from SuPerSim.constants import colorcycle
from SuPerSim.pickling import load_all_pickles

def assign_weight_sim(site, path_pickle, no_weight=True):
    """ Function returns a statistical weight for each simulation according to the importance in rockfall starting zone 
    
    Parameters
    ----------
    site : str
        Location of the event, e.g. 'Aksaut_North', will be used to label all the pickles
    path_pickle : str
        String path to the location of the folder where the pickles are saved
    no_weight : bool, optional
        If True, all simulations have the same weight, otherwise the weight is computed as a function of altitude, aspect, and slope

    Returns
    -------
    pd_weight : pandas.core.frame.DataFrame
        Panda DataFrame assigning a statistical weight to each simulation for each of 'altitude', 'aspect', 'slope'
        and an overall weight.
    pd_weight_long : pandas.core.frame.DataFrame
        Panda DataFrame assigning a statistical weight to each simulation for each of 'altitude', 'aspect', 'slope'
        and an overall weight, and has information about the topography of each simulations
    """

    pkl = load_all_pickles(site, path_pickle)
    df_stats = pkl['df_stats']
    rockfall_values = pkl['rockfall_values']

    dict_weight = {}
    if no_weight:
        dict_weight = {i: [df_stats['altitude'].loc[i], 1, df_stats['aspect'].loc[i], 1, df_stats['slope'].loc[i], 1]
                       for i in list(df_stats.index.values)}
    else:
        if rockfall_values['exact_topo']:
            alt_distance = np.max([np.abs(i-rockfall_values['altitude']) for i in np.sort(np.unique(df_stats['altitude']))])
            dict_weight = {i: [df_stats['altitude'].loc[i],
                            1 - np.abs(df_stats.loc[i]['altitude']-rockfall_values['altitude'])/(2*alt_distance),
                            df_stats['aspect'].loc[i],
                            np.cos((np.pi)/180*(df_stats.loc[i]['aspect']-rockfall_values['aspect']))/4+3/4,
                            df_stats['slope'].loc[i],
                            np.cos((np.pi)/30*(df_stats.loc[i]['slope']-rockfall_values['slope']))/4+3/4]
                            for i in list(df_stats.index.values)}
        else:
            dict_weight = {i: [df_stats['altitude'].loc[i], 1, df_stats['aspect'].loc[i], 1, df_stats['slope'].loc[i], 1]
                        for i in list(df_stats.index.values)}

    pd_weight_long = pd.DataFrame.from_dict(dict_weight, orient='index',
                                        columns=['altitude', 'altitude_weight', 'aspect', 'aspect_weight', 'slope', 'slope_weight'])
    
    pd_weight_long['weight'] = pd_weight_long['altitude_weight']*pd_weight_long['aspect_weight']*pd_weight_long['slope_weight']
    pd_weight = pd_weight_long.drop(columns=['altitude', 'aspect', 'slope'])

    return pd_weight, pd_weight_long

def count_stat_weights(site, path_pickle, no_weight=True): 
    """ Function returns a binned count of the weight distribution over all (valid) simulations 
    
    Parameters
    ----------
    site : str
        Location of the event, e.g. 'Aksaut_North', will be used to label all the pickles
    path_pickle : str
        String path to the location of the folder where the pickles are saved
    no_weight : bool, optional
        If True, all simulations have the same weight, otherwise the weight is computed as a function of altitude, aspect, and slope

    Returns
    -------
    bins : numpy.ndarray
        bins for histogram within the weight disctibution comprised between 0 and 1
    counts : numpy.ndarray
        number of simulations with a weight within each bin
    number_glaciers : int
        number of simulations yielding a 'glacier'
    """

    pd_weight, _ = assign_weight_sim(site, path_pickle, no_weight)

    pkl = load_all_pickles(site, path_pickle)
    df = pkl['df']

    list_hist = pd_weight.loc[:,'weight']
    counts, bins = np.histogram(list_hist, 10, (0, 1))
    
    number_glaciers = len(df)-len(pd_weight)
    
    return bins, counts, number_glaciers

def plot_hist_stat_weights(bins, counts, number_glaciers, show_glaciers=True): 
    """ Function returns a histogram of the weight distribution over all (valid) simulations 
        given a binned count of simulation weights and a number of glaciers
    
    Parameters
    ----------
    bins : numpy.ndarray
        bins for histogram within the weight disctibution comprised between 0 and 1
    counts : numpy.ndarray
        number of simulations with a weight within each bin
    number_glaciers : int
        number of simulations yielding a 'glacier'
    show_glaciers : bool, optional
        If True, shows the glacier simulations with a 0 weight, if False, those are ignored.

    Returns
    -------
    fig : figure
        Histogram
    """

    fig, _ = plt.subplots()

    tot_count = np.sum(counts) + (number_glaciers if show_glaciers else 0)
    counts_glaciers = [0 for _ in counts]
    counts_glaciers[0] = number_glaciers
    counts_all = counts + (counts_glaciers if show_glaciers else 0)
    
    plt.hist(bins[:-1], bins, weights=counts/tot_count, label='No glaciers', color=colorcycle[1])
    if show_glaciers:
        plt.hist(bins[:-1], bins, weights=counts_glaciers/tot_count, label='Glaciers', color=colorcycle[0])

    max_count = np.ceil((np.max(counts_all)/tot_count)/0.05+1)*0.05
    
    ticks = list(np.arange(0, max_count, 0.05))
    plt.yticks(ticks, [f"{i:0.2f}" for i in ticks])

    # Show the graph
    if show_glaciers:
        plt.legend(loc='upper right')
    plt.xlabel('Statistical weight')
    plt.ylabel('Frequency')
    plt.show()
    plt.close()

    return fig

def plot_hist_stat_weights_from_input(site, path_pickle, no_weight=True, show_glaciers=True): 
    """ Function returns a histogram of the weight distribution over all (valid) simulations 
    
    Parameters
    ----------
    site : str
        Location of the event, e.g. 'Aksaut_North', will be used to label all the pickles
    path_pickle : str
        String path to the location of the folder where the pickles are saved
    no_weight : bool, optional
        If True, all simulations have the same weight, otherwise the weight is computed as a function of altitude, aspect, and slope
    show_glaciers : bool, optional
        If True, shows the glacier simulations with a 0 weight, if False, those are ignored.

    Returns
    -------
    fig : figure
        Histogram
    """

    bins, counts, number_glaciers = count_stat_weights(site, path_pickle, no_weight)
    fig = plot_hist_stat_weights(bins, counts, number_glaciers, show_glaciers)

    return fig

def count_perma_sim_per_variable(site, path_thaw_depth, path_pickle): 
    """ Function returns a binned count over all simulations
        of the number of valid/glacier simulations for each of the following variable
        ('altitude', 'aspect', 'slope', 'forcing') 
        It also shows the breakdown of valid simulations into permafrost and no-permafrost ones

    Parameters
    ----------
    site : str
        Location of the event, e.g. 'Aksaut_North', will be used to label all the pickles
    path_thaw_depth : str
        Path to the .nc file where the aggregated thaw depth simulations are stored
    path_pickle : str
        String path to the location of the folder where the pickles are saved

    Returns
    -------
    dict_bins : dict
        Dictionary giving the total number of simulation per variable
        e.g. dict_bins = {'altitude': [2900, 3100, 3300], 'aspect': [22.5, 45.0, 67.5], 'slope': [55, 60, 65, 70, 75], 'forcing': ['merra2']}
    dict_counts : dict
        Dictionary giving the number of simulation per variable and per parmafrost state
        e.g. dict_counts = {'altitude': {'Permafrost': [14, 37, 45], 'No permafrost, no glaciers': [31, 8, 0], 'Glaciers': [0, 0, 0]}, 'aspect': {}, ...}
    """

    _, thaw_depth = open_thaw_depth_nc(path_thaw_depth)

    pkl = load_all_pickles(site, path_pickle)
    df = pkl['df']
    df_stats = pkl['df_stats']

    variables = ['altitude','aspect','slope','forcing']

    list_valid_sim = list(df_stats.index.values)

    list_no_perma = []
    for sim in list_valid_sim:
        if np.std(thaw_depth[sim,:]) < 1 and np.max(thaw_depth[sim,:])> 19:
            list_no_perma.append(sim)

    list_perma = list(set(list_valid_sim) - set(list_no_perma))

    dict_counts = {var: [] for var in variables}
    dict_bins = {var: [] for var in variables}

    for var in variables:
        # number_no_glaciers: number of valid simulations (without glaciers) per value in list_var_prev
        _, number_no_glaciers = [list(i) for i in np.unique(df_stats.loc[:, var], return_counts=True)]
        _, number_perma = [list(i) for i in np.unique((df_stats.loc[list_perma, :]).loc[:, var], return_counts=True)]
        
        # total number of simulations per value per variable
        dict_bins[var], tot = [list(i) for i in np.unique(df.loc[:, var], return_counts=True)]

        if len(number_perma) == 0:
            number_perma = [0 for _ in tot]

        # number_glaciers: number of glaciers per value per variable
        number_glaciers = [tot[i] - number_no_glaciers[i] for i in range(len(tot))]
        # number_no_perma: number of simulation swith no glaciers and no permafrost per value per variable
        number_no_perma = [number_no_glaciers[i] - number_perma[i] for i in range(len(tot))]

        counts = {
            'Permafrost': number_perma,
            'No permafrost, no glaciers': number_no_perma,
            'Glaciers': number_glaciers
        }

        dict_counts[var] = counts

    return dict_bins, dict_counts

def plot_hist_valid_sim_all_variables(dict_bins, dict_counts): 
    """ Function returns a histogram of the number of valid/glacier simulations for each of the following variable
        ('altitude', 'aspect', 'slope', 'forcing') 
        It also shows the breakdown of valid simulations into permafrost and no-permafrost ones

    Parameters
    ----------
    dict_bins : dict
        Dictionary giving the total number of simulation per variable
        e.g. dict_bins = {'altitude': [2900, 3100, 3300], 'aspect': [22.5, 45.0, 67.5], 'slope': [55, 60, 65, 70, 75], 'forcing': ['merra2']}
    dict_counts : dict
        Dictionary giving the number of simulation per variable and per parmafrost state
        e.g. dict_counts = {'altitude': {'Permafrost': [14, 37, 45], 'No permafrost, no glaciers': [31, 8, 0], 'Glaciers': [0, 0, 0]}, 'aspect': {}, ...}

    Returns
    -------
    fig : figure
        Histogram (subplot(2,2))
    """

    variables = ['altitude','aspect','slope','forcing']
    xaxes = ['Altitude [m]','Aspect [°]','Slope [°]','Forcing']
    yaxes = ['Number of simulations','','Number of simulations','']

    tot_per_bin = {var: np.sum([i[0] for i in dict_counts[var].values()]) for var in variables}

    fig, axs = plt.subplots(2,2, figsize=(6,6))
    for idx,ax in enumerate(axs.ravel()):
        bottom = np.zeros(len(dict_bins[variables[idx]]))

        colorbar = {
            'Permafrost': colorcycle[1],
            'No permafrost, no glaciers': colorcycle[2],
            'Glaciers': colorcycle[0]
        }

        for name, data in dict_counts[variables[idx]].items():
            p = ax.bar([str(i) for i in dict_bins[variables[idx]]], data, label=name, bottom=bottom, color=colorbar[name])
            bottom += data
            data_no_zero = [i if i>0 else "" for i in data]
            ax.bar_label(p, labels=data_no_zero, label_type='center')

        ax.set_xlabel(xaxes[idx])
        ax.set_ylim(0, tot_per_bin[variables[idx]])
        ax.set_ylabel(yaxes[idx])

    fig.align_ylabels(axs[:,0])
    plt.legend(loc='lower right', reverse=True)
    plt.tight_layout()
    plt.show()
    plt.close()

    return fig

def plot_hist_valid_sim_all_variables_from_input(site, path_thaw_depth, path_pickle): 
    """ Function returns a binned count over all simulations
        of the number of valid/glacier simulations for each of the following variable
        ('altitude', 'aspect', 'slope', 'forcing') 
        It also shows the breakdown of valid simulations into permafrost and no-permafrost ones

    Parameters
    ----------
    site : str
        Location of the event, e.g. 'Aksaut_North', will be used to label all the pickles
    path_thaw_depth : str
        Path to the .nc file where the aggregated thaw depth simulations are stored
    path_pickle : str
        String path to the location of the folder where the pickles are saved 

    Returns
    -------
    fig : figure
        Histogram (subplot(2,2))
    """

    dict_bins, dict_counts = count_perma_sim_per_variable(site, path_thaw_depth, path_pickle)
    fig = plot_hist_valid_sim_all_variables(dict_bins, dict_counts)

    return fig



"""This module defines the functions that sort the timestamps in bins (years, months, etc.)"""

#pylint: disable=line-too-long

from datetime import datetime
import re
from numpy import ma
import numpy as np
from netCDF4 import num2date #pylint: disable=no-name-in-module

def time_unit_stamp(time_file):
    """ Function returns frequency of datapoints and exact date and time
    of the initial datapoint, converted into a datetime
    
    Parameters
    ----------
    time_file : netCDF4._netCDF4.Variable
        File where the time index of each datapoint is stored (time_ground, not time_air)
        Needs to have a .units attribute

    Returns
    -------
    frequency : str
        Frequency of the data sampling, e.g. 'seconds'
    start_date : datetime.datetime
        Date and time of the first time stamp

    """

    # this extracts the unit of the time series, i.e. when it starts,
    # e.g. days since 0001-1-1 0:0:0 or seconds since 1900-01-01
    # these are two formats that exist and we have to accomodate for both
    # we partition the string and only keep what comes after 'since ': the date
    time_start_date = time_file.units.partition(' since ')[2]

    # we identify the different delimiters between year, month, day, hour, etc.
    # to be: '-', ' ', or ':'
    # we get a list in the form [1, 1, 1, 0, 0, 0] or [1900, 1, 1] respectively
    time_start_date_list = list(map(int, map(float, re.split('-| |:', time_start_date))))

    # now we decide to fill the list with 0s if not long enough
    # i.e. if not provided, we assume the start is at exactly midnight 0:0:0
    time_start_date_fill = time_start_date_list[:6] + [0]*(6 - len(time_start_date_list))

    # here we get the string telling us with what frequency data is sampled, e.g. 'days' or 'seconds'
    frequency = time_file.units.partition(' since ')[0]
    # and the date of the first time stamp
    start_date = datetime(*time_start_date_fill)

    # finally, we convert it to a datetime
    return start_date, frequency

def list_tokens_year(time_file, year_bkg_end, year_trans_end):
    """ Function returns a list of time stamps per year, and splits the timestamps into background and transient periods
    
    Parameters
    ----------
    time_file : netCDF4._netCDF4.Variable
        File where the time index of each datapoint is stored (time_ground or time_air)
        Needs to have a .units attribute
    year_bkg_end : int
        Background period is BEFORE the start of the year corresponding to the variable, i.e. all time stamps before Jan 1st year_bkg_end
    year_trans_end : int
        Same for transient period

    Returns
    -------
    list_years : dict
        a list of indices per calendar year that can be used to study data year per year
        in the form of a dictionary like {1980: [0, 1, 2, ..., 363, 364], 1981: [365, ..., 729], ..., 2019: [14244, ..., 14608]}
    time_bkg : numpy.ma.core.MaskedArray
        Mask that selects only the background timestamps
    time_trans : numpy.ma.core.MaskedArray
        Mask that selects only the transient timestamps
    time_pre_trans : numpy.ma.core.MaskedArray
        Mask that selects all points before the end of the transient period (background+transient)

    """

    # we start by creating a short dictionary that allows us to convert the seconds into the unit of the file
    # e.g. 'days' or 'hours'
    # difference in time stamp between two consecutive measurements (could be 1, 3600, etc.)
    consecutive = ma.getdata(time_file[1]-time_file[0])
    # here we get the unit and we express it in seconds
    seconds_divider = {'seconds': 1, 'minutes': 60, 'hours': 3600, 'days': 86400}[time_unit_stamp(time_file)[1]]
    # we are now able to understand how many seconds there are between 2 consecutive measurements
    secs_between_cons = seconds_divider*consecutive

    # we extract the year and month the data starts at
    start_year = num2date(time_file[0], time_file.units).year
    # start_month = num2date(time_file[0], time_file.units).month
    start_day = num2date(time_file[0], time_file.units).day

    # we extract the exact moment the transient era ends
    # time_end_pre_trans = num2date(time_file[time_pre_trans_file][-1], time_file.units)
    time_end = num2date(time_file[-1], time_file.units)

    # we create a dictionary that associates the correct index position to the start of each new year
    # e.g. if the data is taken daily, all values will be spread by ~365
    list_init = [0 for i in range(year_trans_end-start_year+1)]
    for i in range(year_trans_end-start_year+1):
        # note that we allow an arbitrary 10 days wiggle room just to make sure we capture the end date of the last year
        # of the transient era since technically this last year doesn't go all the way until the new year

        ##################################################################################
        # IT IS VERY IMPORTANT TO HIGHLIGHT THAT THE JRA55 and MERRA2 SCALED DATA DO NOT #
        # SCALE BETWEEN DEC31 6PM AND MIDNIGHT, HENCE LEAVING A 5H GAP IN THE HOURLY     #
        # DATA EVERY YEAR! THIS IS ACCOUNTED FOR BY SUBTRACTING 5 EVERY YEAR BUT ONE HAS #
        # TO BE VERY CAREFUL IF THIS CHANGES SOMEHOW                                     #
        # A SHORT TEST IS IMPLEMENTED AND SHOULD RETURN AN ERROR MESSAGE IF NOT WORKING  #
        ##################################################################################
        if i==0:
            list_init[i] = 0
        else:
            # time difference between yy-01-01 00:00:00 and exactly a year before
            dt = int((datetime(start_year+i,1,1,0,0,0)
                    - datetime(start_year+i-1,1,(start_day if i==1 else 1),0,0,0)).total_seconds()/secs_between_cons)
            # previous entry in the catalogue, in days elapsed from the start
            prev_year = list_init[i-1]
            # previous year + dt should correspond to Jan 1st at 00:00:00 of the current year
            list_init[i] = prev_year + dt
            # this is where we check that there is no gap in the data by making sure current_year
            # indeed corresponds to Jan 1st at 00:00:00 of the current year
            check = int((num2date(time_file[list_init[i]], time_file.units) - datetime(start_year+i,1,1,0,0,0)).total_seconds())
            # finally, we assign the time in days from the start for this month if the check is correct.
            # however, if the check is not, this means that we are in the situation where there is 5 data
            # point missing between 6pm and midnight on December 31st for either merra2 or jra55
            list_init[i] += (0 if check ==0 else -5)

    last = int((time_end - num2date(time_file[0], time_file.units)).total_seconds()/secs_between_cons)

    # finally, we have a dictionary that associates each year in the pre-transient era to a list of
    # indices in the time series corresponding to that particular year
    list_years = {i+start_year: list(range(list_init[i], (list_init[i+1] if i+1+start_year<=year_trans_end else last+1)))
                for i in range(len(list_init))}

    # this function selects all data before the background cutoff, hence selects the background data
    time_bkg = np.less(time_file[:], time_file[list_years[year_bkg_end][0]])
    # this function selects all data after the background cutoff, but before the transient one, hence selects the transient data
    time_trans = np.logical_and(time_file[:] >= time_file[list_years[year_bkg_end][0]], time_file[:] < time_file[list_years[year_trans_end][0]])
    # this function selects all data before the transient cutoff, hence selects the pre-transient data
    time_pre_trans = np.less(time_file[:], time_file[list_years[year_trans_end][0]])


    # It is here that we make sure that the data that should correspond to Jan 1, 0:00:00 of the last year
    # actually does, otherwise it raises an error and stops the evaluation
    diff = int((num2date(time_file[list_years[year_trans_end][0]], time_file.units)-
                datetime(year_trans_end,1,1,0,0,0)).total_seconds())
    if diff!=0:
        raise ValueError('VictorCustomError: scaled JRA55 and MERRA2 have unidentified gaps in data, check script')

    return list_years, time_bkg, time_trans, time_pre_trans

def specific_time_to_index(time_file, date):
    """ Function returns the timeseries time index corresponding to a given date
    
    Parameters
    ----------
    time_file : netCDF4._netCDF4.Variable
        File where the time index of each datapoint is stored (time_ground, not time_air)
        Needs to have a .units attribute
    date : datetime.datetime
        Datetime of the event, e.g. datetime(2019, 5, 13, 0, 0)

    Returns
    -------
    time_stamp : int
        Time stamp of the particular date for the timeseries
    """

    delta_cons_time = int((num2date(time_file[1], time_file.units) - num2date(time_file[0], time_file.units)).total_seconds())
    init_delta = int((date - num2date(time_file[0], time_file.units)).total_seconds())
    time_stamp = int(init_delta/delta_cons_time)
    refined_delta = int((date - num2date(time_file[time_stamp], time_file.units)).total_seconds())
    while refined_delta < 0:
        time_stamp -= 1
        refined_delta = int((date - num2date(time_file[time_stamp], time_file.units)).total_seconds())
    while refined_delta > 0:
        time_stamp += 1
        refined_delta = int((date - num2date(time_file[time_stamp], time_file.units)).total_seconds())

    return time_stamp



"""This module defines the functions that create statistics for the timseries and pickle them"""

#pylint: disable=line-too-long
#pylint: disable=trailing-whitespace
#pylint: disable=invalid-name

import pickle
from collections import defaultdict
from datetime import datetime
import numpy as np
import pandas as pd

from SuPerSim.open import open_air_nc, open_ground_nc, open_snow_nc
from SuPerSim.mytime import list_tokens_year, specific_time_to_index

def assign_value_global_dict(path_forcing_list, path_ground, path_snow, path_pickle, year_bkg_end, year_trans_end, site):
    """ Function returns a dictionary containing all the important timeseries and saves it to a pickle
    
    Parameters
    ----------
    path_forcing_list : list of str
        List of paths to the .nc file where the atmospheric forcing data for the given reanalysis is stored
    path_ground : str
        Path to the .nc file where the aggregated ground simulations are stored
    path_snow : str
        Path to the .nc file where the aggregated snow simulations are stored
    path_pickle : str
        String path to the location of the folder where the pickles are saved
    year_bkg_end : int
        Background period is BEFORE the start of the year corresponding to the variable, i.e. all time stamps before Jan 1st year_bkg_end
    year_trans_end : int
        Same for transient period
    site : str
        Location of the event, e.g. 'Aksaut_North', will be used to label all the pickles   


    Returns
    -------
    global_dict : dict
        A large dictionary listing all timeseries, organized in: air/ground/snow > relevant timeseries
        For the 'air', all timeseries are a list with each entry corresponding to a different reanalysis

    """

    file_name = f"global_dict{('' if site=='' else '_')}{site}.pkl"
    my_path = path_pickle + file_name

    # try to open the pickle file, if it exists
    try:
        # Open the file in binary mode
        with open(my_path, 'rb') as file:
            # Call load method to deserialze
            global_dict = pickle.load(file)
        print('Succesfully opened the pre-existing pickle:', file_name)

    # if the pickle file does not exist, we have to create it
    except (OSError, IOError) as e:
        list_vars = ['time_air', 'temp_air', 'SW_flux', 'SW_direct_flux', 'SW_diffuse_flux', 'precipitation',
                     'time_air_bkg', 'time_air_trans', 'time_air_pre_trans']

        list_series = [open_air_nc(i) + list_tokens_year(open_air_nc(i)[0], year_bkg_end, year_trans_end)[1:] for i in path_forcing_list]
        list_series_b = [[list_series[j][i] for j in range(len(list_series))] for i in range(len(list_series[0]))]

        global_dict = {'air': dict(zip(list_vars, list_series_b)),
                    'ground': dict(zip(['time_ground', 'temp_ground'], open_ground_nc(path_ground)[1:])),
                    'snow': dict(zip(['snow_height'], open_snow_nc(path_snow)[1:]))}
        
        # Open a file and use dump()
        with open(my_path, 'wb') as file:
            # A new file will be created
            pickle.dump(global_dict, file)
        print('Created a new pickle:', file_name)

        # useless line just to use the variable 'e' so that I don't get an error
        if e == 0:
            pass

    return global_dict

def assign_value_df_raw(path_repository):
    """ Function converts the .csv ensemble simulation repository into a panda dataframe with a column for each parameter
    
    Parameters
    ----------
    path_repository : str
        Path to the .csv file with all the simulation parameters

    Returns
    -------
    df_raw : pandas.core.frame.DataFrame
        A panda dataframe version of the .csv file where the simulation paramaters have been unpacked into readable columns

    """

    df_raw = pd.read_csv(path_repository, usecols=['site','directory','parameters'])
    df_raw['altitude'] = (df_raw['site'].str.split('_').str[1]).apply(pd.to_numeric)
    df_raw['site_name'] = df_raw['site'].str.split('_').str[0]
    df_raw['forcing'] = df_raw['directory'].str.split('_').str[3]
    df_raw['aspect'] = [pd.to_numeric(i.replace('p','.')) for i in (df_raw['parameters'].str.split('aspect_').str[1]).str.split('.inpts').str[0]]
    df_raw['slope'] = ((df_raw['parameters'].str.split('slope_').str[1]).str.split('.inpts').str[0]).apply(pd.to_numeric)
    df_raw['snow'] = [(y/100 if y > 10 else y) for y in [pd.to_numeric(i.replace('p','.')) for i in ((df_raw['parameters'].str.split('snow_').str[1]).str.split('.inpts').str[0])]]
    if len(df_raw['parameters'].str.split('pointmaxswe_')[0]) == 2:
        df_raw['maxswe'] = (df_raw['parameters'].str.split('pointmaxswe_').str[1]).str.split('.inpts').str[0]
    else:
        df_raw['maxswe'] = [np.nan for i in df_raw['altitude']]
    df_raw['material'] = (df_raw['parameters'].str.split('soil_').str[1]).str.split('.inpts').str[0]
    df_raw.drop('parameters', axis=1, inplace=True)

    return df_raw

def assign_value_df(path_repository, path_ground, path_pickle, site):
    """ Function returns the panda dataframe with all ensemble simulation parameters and saves it to a pickle
    
    Parameters
    ----------
    path_repository : str
        Path to the .csv file with all the simulation parameters
    path_ground : str
        Path to the .nc file where the aggregated ground simulations are stored
    path_pickle : str
        String path to the location of the folder where the pickles are saved
    site : str
        Location of the event, e.g. 'Aksaut_North', will be used to label all the pickles   


    Returns
    -------
    df : pandas.core.frame.DataFrame
        A panda dataframe version of the .csv file where the simulation paramaters have been unpacked into readable columns,
        saved to a pickle file
        The 'directory' column is ordered by the simulation index provided in 'path_ground'

    """
    # INPUT: Should ONLY take var_name = df
    # OUPUT: value of df whether the variable was already assigned or not.
    # This saves a lot of time since it is not re-evaluated each time but just the first
    # time and then saved into a pickle.

    file_name = f"df{('' if site=='' else '_')}{site}.pkl"
    my_path = path_pickle + file_name

    f_ground, _, _ = open_ground_nc(path_ground)

    # try to open the pickle file, if it exists
    try:
        # Open the file in binary mode
        with open(my_path, 'rb') as file:
            # Call load method to deserialze
            df = pickle.load(file) 
        print('Succesfully opened the pre-existing pickle:', file_name)

    # if the pickle file does not exist, we have to create it
    except (OSError, IOError) as e:
        df_raw = assign_value_df_raw(path_repository)
            
        # this is a method that re-orders the 'directory' column thanks to the simulation index of f_ground
        df_raw.directory = df_raw.directory.astype("category")
        df_raw.directory = df_raw.directory.cat.set_categories(f_ground['simulation'][:])
        # now the order of the csv file corresponds to the order of the ground
        df = df_raw.sort_values(['directory']).reset_index(drop=True)

        if all(f_ground['simulation'][:,:] == df.iloc[:].directory):
            print('All good, same order for all lists!')
        else:
            print('!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! ERROR !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!')
        
        # Open a file and use dump() 
        with open(my_path, 'wb') as file: 
            # A new file will be created 
            pickle.dump(df, file)
        print('Created a new pickle:', file_name)

        # useless line just to use the variable 'e' so that I don't get an error
        if e == 0:
            pass

    return df

def assign_value_reanalysis_stat(forcing_list, path_forcing_list, path_pickle, year_bkg_end, year_trans_end, site):
    """ Creates a dictionary of mean quantities over the background and transient periods
    
    Parameters
    ----------
    forcing_list : list of str
        List of forcings provided, with a number of entries between 1 and 3 in 'era5', 'merra2', and 'jra55'. E.g. ['era5', 'merra2']
    path_forcing_list : list of str
        List of paths to the .nc file where the atmospheric forcing data for the given reanalysis is stored
    path_pickle : str
        String path to the location of the folder where the pickles are saved
    year_bkg_end : int
        Background period is BEFORE the start of the year corresponding to the variable, i.e. all time stamps before Jan 1st year_bkg_end
    year_trans_end : int
        Same for transient period
    site : str
        Location of the event, e.g. 'Aksaut_Ridge' (not a list, 1 single location)    

    Returns
    -------
    reanalysis_stats : dict
        dictionary of mean quntities over the background and transient periods

    """

    file_name = f"reanalysis_stats{('' if site=='' else '_')}{site}.pkl"
    file_name_df = f"df{('' if site=='' else '_')}{site}.pkl"
    my_path = path_pickle + file_name
    my_path_df = path_pickle + file_name_df

    # This saves a lot of time since it is not re-evaluated each time but just the first
    # time and then saved into a pickle.

    # try to open the pickle file, if it exists
    try: 
        # Open the file in binary mode 
        with open(my_path, 'rb') as file: 
            # Call load method to deserialize 
            var_name = pickle.load(file) 
        print('Succesfully opened the pre-existing pickle:', file_name)

    # if the pickle file does not exist, we have to create it
    except (OSError, IOError) as e:
        # difference (trend) between trans and bkg periods for the air temperature:
        # amount of warming, at different altitudes

        temp_air, SW_flux, SW_direct_flux, SW_diffuse_flux, time_bkg_air, time_trans_air, time_pre_trans_air = [[] for _ in range(7)]

        for i in range(len(forcing_list)):
            time_air, var1, var2, var3, var4, var5, var6, var7 = [[] for _ in range(8)]
            time_air, var1, var2, var3, var4, _ = open_air_nc(path_forcing_list[i])
            temp_air.append(var1)
            SW_flux.append(var2)
            SW_direct_flux.append(var3)
            SW_diffuse_flux.append(var4)
            _, var5, var6, var7 = list_tokens_year(time_air, year_bkg_end, year_trans_end)
            time_bkg_air.append(var5)
            time_trans_air.append(var6)
            time_pre_trans_air.append(var7)

        # Open the file in binary mode
        with open(my_path_df, 'rb') as file:
            # Call load method to deserialze
            df = pickle.load(file) 

        # list of sorted altitudes at the site studied
        list_alt = sorted(list(set(df.altitude)))

        var_name = {forcing_list[i]: {list_alt[alt]:
                                {'temp_bkg': temp_air[i][time_bkg_air[i], alt].mean(),
                                    # background air temperature at different elevations
                                    'temp_trans': temp_air[i][time_trans_air[i], alt].mean(),
                                    # transient air temperature at different elevations
                                    'air_warming': temp_air[i][time_trans_air[i], alt].mean()
                                    - temp_air[i][time_bkg_air[i], alt].mean(),
                                    # air warming at different elevations
                                    'SW': SW_flux[i][time_pre_trans_air[i], alt].mean(),
                                    # total flux mean over the whole study, at different elevations
                                    'SW_direct': SW_direct_flux[i][time_pre_trans_air[i], alt].mean(),
                                    # total direct flux mean over the whole study, at different elevations
                                    'SW_diffuse': SW_diffuse_flux[i][time_pre_trans_air[i], alt].mean()
                                    # total diffuse flux mean over the whole study, at different elevations
                                }
                                for alt in range(temp_air[i].shape[1])}
                    for i in range(len(forcing_list))}
        
        # Open a file and use dump() 
        with open(my_path, 'wb') as file: 
            # A new file will be created 
            pickle.dump(var_name, file)
        print('Created a new pickle:', file_name)

        # useless line just to use the variable 'e' so that I don't get an error
        if e == 0:
            pass

    return var_name

def glacier_filter(site, path_snow, path_pickle, glacier=False, min_glacier_depth=100, max_glacier_depth=20000):
    """ Function returns a list of valid simulations regarding the glacier criteria
    
    Parameters
    ----------
    site : str
        Location of the event, e.g. 'Aksaut_Ridge'
    path_snow : str
        Path to the .nc file where the aggregated snow simulations are stored
    path_pickle : str
        String path to the location of the folder where the pickles are saved
    glacier : bool, optional
        By default only keeps non-glacier simulations but can be changed to True to select only glaciated simulations
    min_glacier_depth : float, optional
        Selects simulation with minimum snow height higher than this threshold (in mm)
    max_glacier_depth : float, optional
        Selects simulation with minimum snow height lower than this threshold (in mm)
    

    Returns
    -------
    list_valid_sim : list
        list of simulation number for all valid simulations
    """
    # This saves a lot of time since it is not re-evaluated each time but just the first
    # time and then saved into a pickle.

    _, snow_height = open_snow_nc(path_snow)

    file_name = f"list_valid_sim{('' if site=='' else '_')}{site}.pkl"
    my_path = path_pickle + file_name

    # try to open the pickle file, if it exists
    try: 
        # Open the file in binary mode 
        with open(my_path, 'rb') as file: 
            # Call load method to deserialize 
            list_valid_sim = pickle.load(file) 
        print('Succesfully opened the pre-existing pickle:', file_name)

    # if the pickle file does not exist, we have to create it
    except (OSError, IOError) as e:
        # we create a dictionary of all valid simulations
        list_valid_sim = []
        for sim_index in range(snow_height.shape[0]):
            min_snow_height = np.min(snow_height[sim_index,:])
            if glacier:
                if (min_snow_height >= min_glacier_depth) & (min_snow_height < max_glacier_depth):
                    list_valid_sim.append(sim_index)
                else:
                    pass
            else:
                if min_snow_height < min_glacier_depth:
                    list_valid_sim.append(sim_index)
                else:
                    pass
        
        # Open a file and use dump() 
        with open(my_path, 'wb') as file:
            # A new file will be created 
            pickle.dump(list_valid_sim, file)
        print('Created a new pickle:', file_name)

        # useless line just to use the variable 'e' so that I don't get an error
        if e == 0:
            pass

    return list_valid_sim

def melt_out_date(consecutive, path_ground, path_snow, path_pickle, year_bkg_end, year_trans_end, site):
    """ Function returns a list of melt out dates given the criterion of a number of consecutive snow-free days
    
    Parameters
    ----------
    consecutive : int
        Number of minimum consecutive snow-free days to declare seasonal melt out of the snow
    path_ground : str
        Path to the .nc file where the aggregated ground simulations are stored
    path_snow : str
        Path to the .nc file where the aggregated snow simulations are stored
    path_pickle : str
        String path to the location of the folder where the pickles are saved
    year_bkg_end : int
        Background period is BEFORE the start of the year corresponding to the variable, i.e. all time stamps before Jan 1st year_bkg_end
    year_trans_end : int
        Same for transient period
    site : str
        Location of the event, e.g. 'Aksaut_North', will be used to label all the pickles
    

    Returns
    -------
    dict_melt_out : dict
        Dictionary that assings a melt-out date to each simulation and each year
        Note that if snow does NOT melt at all, we assign the maximal value corresponding to the end of the year 
    stats_melt_out_dic : dict
        Background, transient and full mean of the melt-out date for each simulation
    """

    # This saves a lot of time since it is not re-evaluated each time but just the first
    # time and then saved into a pickle. 

    file_name = f"melt_out{('' if site=='' else '_')}{site}.pkl"
    file_name_list_valid_sim = f"list_valid_sim{('' if site=='' else '_')}{site}.pkl"
    my_path = path_pickle + file_name

    snow_height = open_snow_nc(path_snow)[1]
    time_ground = open_ground_nc(path_ground)[1]
    with open(path_pickle + file_name_list_valid_sim, 'rb') as file: 
        # Call load method to deserialize 
        list_valid_sim = pickle.load(file)

    # try to open the pickle file, if it exists
    try: 
        # Open the file in binary mode 
        with open(my_path, 'rb') as file: 
            # Call load method to deserialze 
            dict_melt_out, stats_melt_out_dic = pickle.load(file) 
        print('Succesfully opened the pre-existing pickle:', file_name)

    # if the pickle file does not exist, we have to create it
    except (OSError, IOError) as e:
        # we start by importing the list of time indexes separated by year
        dictionary = list_tokens_year(time_ground, year_bkg_end, year_trans_end)[0]

        # we create the dictionary of melt-out dates
        dict_melt_out = defaultdict(dict)
        for sim in list_valid_sim:
            for year in list(dictionary.keys()):
                # we extract the time series indices corresponding to the year selected
                time_particular_year = dictionary[year]

                # we get a list of snow free day
                test_list = [i for i, x in enumerate(snow_height[sim, time_particular_year]) if x == 0]
                # this counter goes through test_list and checks that there is a number of consecutive days without snow
                counter = 0
                # this will be the final result and is initialized to 0
                result = 0
                # we start by making sure that the list of snow-free days is at least as long as the number of consecutive
                # days we are after, if not we conclude that there is snow all year round
                if len(test_list) >= counter + consecutive + 1:   
                    if test_list[counter+consecutive] == test_list[counter]+consecutive:
                        result = test_list[counter]
                    # keep going through the loop while we don't have the number of snow-free consecutive days
                    while test_list[counter+consecutive] != test_list[counter]+consecutive:
                        counter += 1
                        result = test_list[counter]
                        if len(test_list) < counter + consecutive + 1:
                            result = len(snow_height[sim, time_particular_year])
                            break
                else:
                    result = len(snow_height[sim, time_particular_year])

                dict_melt_out[sim][year] = result
            
        stats_melt_out_dic = {}
        for sim_index in list(dict_melt_out.keys()):
            stats_melt_out_dic[sim_index] = {'bkg_mean': np.mean([dict_melt_out[sim_index][k] for k in list(dict_melt_out[sim_index].keys()) if k < year_bkg_end]),
                                             'trans_mean': np.mean([dict_melt_out[sim_index][k] for k in list(dict_melt_out[sim_index].keys()) if k >= year_bkg_end]),
                                             'full_mean': np.mean(list(dict_melt_out[sim_index].values())),
                                             'full_std': np.std(list(dict_melt_out[sim_index].values()))}
        
        # Open a file and use dump() 
        with open(my_path, 'wb') as file:
            # A new file will be created 
            pickle.dump([dict_melt_out, stats_melt_out_dic], file)
        print('Created a new pickle:', file_name)

        # useless line just to use the variable 'e' so that I don't get an error
        if e == 0:
            pass

    return dict_melt_out, stats_melt_out_dic

def assign_value_df_stats(path_ground, path_snow, path_pickle, year_bkg_end, year_trans_end, site):
    """ Function returns a large panda dataframe with information about the air, ground, snow, topography, etc. for all simulations
    
    Parameters
    ----------
    path_ground : str
        Path to the .nc file where the aggregated ground simulations are stored
    path_snow : str
        Path to the .nc file where the aggregated snow simulations are stored
    path_pickle : str
        String path to the location of the folder where the pickles are saved
    year_bkg_end : int
        Background period is BEFORE the start of the year corresponding to the variable, i.e. all time stamps before Jan 1st year_bkg_end
    year_trans_end : int
        Same for transient period
    site : str
        Location of the event, e.g. 'Aksaut_North', will be used to label all the pickles
    

    Returns
    -------
    df_stats : pandas.core.frame.DataFrame
        Large panda dataframe with information about the air, ground, snow, topography, etc. for all simulations
    """

    file_name = f"df_stats{('' if site=='' else '_')}{site}.pkl"
    file_name_df = f"df{('' if site=='' else '_')}{site}.pkl"
    file_name_reanalysis_stats = f"reanalysis_stats{('' if site=='' else '_')}{site}.pkl"
    file_name_list_valid_sim = f"list_valid_sim{('' if site=='' else '_')}{site}.pkl"
    file_name_melt_out = f"melt_out{('' if site=='' else '_')}{site}.pkl"
    my_path = path_pickle + file_name

    snow_height = open_snow_nc(path_snow)[1]
    _, time_ground, temp_ground = open_ground_nc(path_ground)
    _, time_bkg_ground, time_trans_ground, time_pre_trans_ground = list_tokens_year(time_ground, year_bkg_end, year_trans_end)

    with open(path_pickle + file_name_df, 'rb') as file: 
        # Call load method to deserialize 
        df = pickle.load(file)

    with open(path_pickle + file_name_reanalysis_stats, 'rb') as file: 
        # Call load method to deserialize 
        reanalysis_stats = pickle.load(file)

    with open(path_pickle + file_name_list_valid_sim, 'rb') as file: 
        # Call load method to deserialize 
        list_valid_sim = pickle.load(file)

    with open(path_pickle + file_name_melt_out, 'rb') as file: 
        # Call load method to deserialize 
        _, stats_melt_out_dic = pickle.load(file)

    # try to open the pickle file, if it exists
    try: 
        # Open the file in binary mode 
        with open(my_path, 'rb') as file: 
            # Call load method to deserialze 
            var_name = pickle.load(file) 
        print('Succesfully opened the pre-existing pickle:', file_name)

    # if the pickle file does not exist, we have to create it
    except (OSError, IOError) as e:
        # this dictionary will tell you pretty much everything you need to know for each simulation
        dict_stats = {}
        len_tot = list(time_pre_trans_ground).count(True)
        len_bkg = list(time_bkg_ground).count(True)
        len_trans = list(time_trans_ground).count(True)

        for sim_index in list_valid_sim:
            dict_stats[sim_index] = [df.iloc[sim_index].directory,
                                    # hash code of the simulation, e.g. 'gt_joffre_2000_merra2_ba887ec'
                                    temp_ground[sim_index, time_bkg_ground,0].mean(),
                                    # background ground temperature for that particular simulation
                                    temp_ground[sim_index, time_trans_ground,0].mean(),
                                    # transient ground temperature for that particular simulation
                                    temp_ground[sim_index, time_trans_ground,0].mean()
                                    - temp_ground[sim_index, time_bkg_ground,0].mean(),
                                    # ground temperature evolution between background and transient for that particular simulation
                                    reanalysis_stats[df.loc[sim_index].forcing][df.loc[sim_index].altitude]['temp_bkg'],
                                    # background air temperature for that particular simulation
                                    reanalysis_stats[df.loc[sim_index].forcing][df.loc[sim_index].altitude]['temp_trans'],
                                    # transient air temperature for that particular simulation
                                    reanalysis_stats[df.loc[sim_index].forcing][df.loc[sim_index].altitude]['air_warming'],
                                    # air temperature evolution between background and transient for that particular simulation
                                    temp_ground[sim_index, time_bkg_ground,0].mean()
                                    - reanalysis_stats[df.iloc[sim_index].forcing][df.iloc[sim_index].altitude]['temp_bkg'],
                                    # background surface offset (SO) for that particular simulation
                                    temp_ground[sim_index, time_trans_ground,0].mean()
                                    - reanalysis_stats[df.iloc[sim_index].forcing][df.iloc[sim_index].altitude]['temp_trans'],
                                    # transient surface offset (SO) for that particular simulation
                                    temp_ground[sim_index, time_trans_ground,0].mean()
                                    - temp_ground[sim_index, time_bkg_ground,0].mean()
                                    - reanalysis_stats[df.iloc[sim_index].forcing][df.iloc[sim_index].altitude]['air_warming'],
                                    # differential warming for that particular simulation
                                    temp_ground[sim_index, time_pre_trans_ground, 0].std(),
                                    # standard deviation of the temperature for that particular simulation
                                    np.min(snow_height[sim_index, time_pre_trans_ground]),
                                    # minimum snow depth
                                    np.max(snow_height[sim_index, time_pre_trans_ground]),
                                    # maximum snow depth 
                                    snow_height[sim_index, time_pre_trans_ground].mean(),
                                    # average snow depth
                                    snow_height[sim_index, time_bkg_ground].mean(),
                                    # average snow depth for the background
                                    snow_height[sim_index, time_trans_ground].mean(),
                                    # average snow depth for the transient
                                    stats_melt_out_dic[sim_index]['full_mean'],
                                    # average snow melt out date
                                    stats_melt_out_dic[sim_index]['full_std'],
                                    # standard deviation of the snow melt out date
                                    stats_melt_out_dic[sim_index]['bkg_mean'],
                                    # average snow melt out date for the background
                                    stats_melt_out_dic[sim_index]['trans_mean'],
                                    # average snow melt out date for the transient
                                    np.count_nonzero(snow_height[sim_index, time_pre_trans_ground])/len_tot,
                                    # fraction of days with snow
                                    np.count_nonzero(snow_height[sim_index, time_bkg_ground])/len_bkg,
                                    # fraction of days with snow for the background
                                    np.count_nonzero(snow_height[sim_index, time_trans_ground])/len_trans,
                                    # fraction of days with snow for the transient
                                    reanalysis_stats[df.loc[sim_index].forcing][df.loc[sim_index].altitude]['SW'],
                                    # mean SW for that particular simulation
                                    reanalysis_stats[df.loc[sim_index].forcing][df.loc[sim_index].altitude]['SW_direct'],
                                    # mean direct SW for that particular simulation
                                    reanalysis_stats[df.loc[sim_index].forcing][df.loc[sim_index].altitude]['SW_diffuse'],
                                    # mean diffuse SW for that particular simulation
                                    df.iloc[sim_index].altitude, # altitude of the simulation
                                    df.iloc[sim_index].aspect, # aspect of the simulation
                                    df.iloc[sim_index].slope, # slope of the simulation
                                    df.iloc[sim_index].snow, # snow correction factor of the simulation
                                    df.iloc[sim_index].maxswe, # maximum snow water equivalent of the simulation
                                    df.iloc[sim_index].material, # ground material of the simulation, e.g. 'rock'
                                    df.iloc[sim_index].forcing, # data reanalysis forcing of the simulation, e.g. 'era5'
                                    df.iloc[sim_index].site_name, # site name of the simulation
                                    ]
        
        var_name = pd.DataFrame.from_dict(dict_stats,
                                            orient='index',
                                            columns=['directory', 'bkg_grd_temp', 'trans_grd_temp', 'evol_grd_temp',
                                                    'bkg_air_temp', 'trans_air_temp', 'evol_air_temp',
                                                    'bkg_SO', 'trans_SO', 'diff_warming', 'std_temp',
                                                    'min_snow', 'max_snow', 'mean_snow', 'mean_snow_bkg', 'mean_snow_trans',
                                                    'melt_out_mean', 'melt_out_std', 'melt_out_bkg', 'melt_out_trans',
                                                    'frac_snow', 'frac_snow_bkg', 'frac_snow_trans',
                                                    'mean_SW','mean_SW_direct','mean_SW_diffuse', 'altitude', 'aspect',
                                                    'slope', 'snow_corr', 'max_swe', 'material', 'forcing', 'site_name'])
        
        # Open a file and use dump() 
        with open(my_path, 'wb') as file:
            # A new file will be created 
            pickle.dump(var_name, file)
        print('Created a new pickle:', file_name)

        # useless line just to use the variable 'e' so that I don't get an error
        if e == 0:
            pass

    return var_name

def assign_rockfall_values(site, path_pickle, path_ground, path_forcing_list, date_event, topo_event):
    """ Function returns a dictionary with date and topography information for the event
    
    Parameters
    ----------
    path_ground : str
        Path to the .nc file where the aggregated ground simulations are stored
    path_pickle : str
        String path to the location of the folder where the pickles are saved
    path_forcing_list : list of str
        List of paths to the .nc file where the atmospheric forcing data for the given reanalysis is stored
    date_event : list
        Date of the event in the form [yy, mm, dd]. If only partial information are available, [yy] works too.
        If no information, set date_event = []
    topo_event : list
        Topography of the event in the form [aspect, slope, altitude].
        If no information, set topo_event = []
    site : str
        Location of the event, e.g. 'Aksaut_North', will be used to label all the pickles   

    Returns
    -------
    rockfall_values : dict
        Dictionary with date and topography info for the event
    """

    file_name = f"rockfall_values{('' if site=='' else '_')}{site}.pkl"
    my_path = path_pickle + file_name

    # try to open the pickle file, if it exists
    try:
        # Open the file in binary mode
        with open(my_path, 'rb') as file:
            # Call load method to deserialze
            rockfall_values = pickle.load(file) 
        print('Succesfully opened the pre-existing pickle:', file_name)

    # if the pickle file does not exist, we have to create it
    except (OSError, IOError) as e:
        _, time_ground, _ = open_ground_nc(path_ground)
        time_air_all = [open_air_nc(i)[0] for i in path_forcing_list]

        rockfall_values = {}
        if [isinstance(i, int) for i in date_event] == [True, True, True]:
            rockfall_values['exact_date'] = True
            rockfall_values['year'] = date_event[0] if isinstance(date_event[0], int) else 0
            rockfall_values['datetime'] = datetime(date_event[0], date_event[1], date_event[2], 0, 0, 0, 0)

            time_index = []
        
            for i in [time_ground, *time_air_all]:
                time_index.append(specific_time_to_index(i, rockfall_values['datetime']))

            time_index = list(np.unique(time_index))

            rockfall_values['time_index'] = time_index

        else:
            rockfall_values['exact_date'] = False
            if len(date_event) > 0:
                if isinstance(date_event[0], int):
                    rockfall_values['year'] = date_event[0]
        if [type(i) for i in topo_event] == [int, int, int]:
            rockfall_values['exact_topo'] = True
            rockfall_values['aspect'], rockfall_values['slope'], rockfall_values['altitude'] = topo_event
        else:
            rockfall_values['exact_topo'] = False
        
        # Open a file and use dump() 
        with open(my_path, 'wb') as file: 
            # A new file will be created 
            pickle.dump(rockfall_values, file)
        print('Created a new pickle:', file_name)

        # useless line just to use the variable 'e' so that I don't get an error
        if e == 0:
            pass

    return rockfall_values

def load_all_pickles(site, path_pickle):
    """ Loads all pickles corresponding to the site name
    
    Parameters
    ----------
    site : str
        Location of the event, e.g. 'Aksaut_Ridge'
    path_pickle : str
        String path to the location of the folder where the pickles are saved

    Returns
    -------
    pkl : dict
        Dictionary containing all the pickles with keys:
        {df, reanalysis_stats, list_valid_sim, dict_melt_out, stats_melt_out_dic, df_stats, rockfall_values}
        Each pickle is defined by
        df : pandas.core.frame.DataFrame
            A panda dataframe version of the .csv file where the simulation paramaters have been unpacked into readable columns,
            saved to a pickle file
            The 'directory' column is ordered by the simulation index provided in 'path_ground'
        reanalysis_stats : dict
            dictionary of mean quntities over the background and transient periods
        list_valid_sim : list
            list of simulation number for all valid simulations
        dict_melt_out : dict
            Dictionary that assings a melt-out date to each simulation and each year
            Note that if snow does NOT melt at all, we assign the maximal value corresponding to the end of the year 
        stats_melt_out_dic : dict
            Background, transient and full mean of the melt-out date for each simulation
        df_stats : pandas.core.frame.DataFrame
            Large panda dataframe with information about the air, ground, snow, topography, etc. for all simulations
        rockfall_values : dict
            Dictionary with date and topography info for the event

    """

    list_file_names = [f"df{('' if site=='' else '_')}{site}.pkl",
                       f"reanalysis_stats{('' if site=='' else '_')}{site}.pkl",
                       f"list_valid_sim{('' if site=='' else '_')}{site}.pkl",
                       f"melt_out{('' if site=='' else '_')}{site}.pkl",
                       f"df_stats{('' if site=='' else '_')}{site}.pkl",
                       f"rockfall_values{('' if site=='' else '_')}{site}.pkl"]

    output = [0 for _ in list_file_names]

    for i, file_name in enumerate(list_file_names):
        my_path = path_pickle + file_name
        # Open the file in binary mode 
        with open(my_path, 'rb') as file: 
            # Call load method to deserialze 
            output[i] = pickle.load(file) 

    [df, reanalysis_stats, list_valid_sim, [dict_melt_out, stats_melt_out_dic], df_stats, rockfall_values] = output

    pkl = {'df': df,
           'reanalysis_stats': reanalysis_stats,
           'list_valid_sim': list_valid_sim,
           'dict_melt_out': dict_melt_out,
           'stats_melt_out_dic': stats_melt_out_dic,
           'df_stats': df_stats,
           'rockfall_values': rockfall_values}

    return pkl

def get_all_stats(forcing_list, path_forcing_list, path_repository, path_ground, path_snow, path_pickle,
                  year_bkg_end, year_trans_end, consecutive,
                  site, date_event, topo_event,
                  glacier=False, min_glacier_depth=100, max_glacier_depth=20000):
    """ Creates a number of pickle files (if they don't exist yet)
    
    Parameters
    ----------
    forcing_list : list of str
        List of forcings provided, with a number of entries between 1 and 3 in 'era5', 'merra2', and 'jra55'. E.g. ['era5', 'merra2']
    path_forcing_list : list of str
        List of paths to the .nc file where the atmospheric forcing data for the given reanalysis is stored
    path_repository : str
        Path to the .csv file with all the simulation parameters
    path_ground : str
        Path to the .nc file where the aggregated ground simulations are stored
    path_snow : str
        path to the snow data
    path_pickle : str
        String path to the location of the folder where the pickles are saved
    year_bkg_end : int
        Background period is BEFORE the start of the year corresponding to the variable, i.e. all time stamps before Jan 1st year_bkg_end
    year_trans_end : int
        Same for transient period
    consecutive : int
        number of consecutive snow-free days to consider that snow has melted for the season
    site : str
        Location of the event, e.g. 'Aksaut_Ridge' (not a list, 1 single location)
    date_event : list
        Date of the event in the form [yy, mm, dd]. If only partial information are available, [yy] works too.
        If no information, set date_event = []
    topo_event : list
        Topography of the event in the form [aspect, slope, altitude].
        If no information, set topo_event = []
    glacier : bool, optional
        By default only keeps non-glacier simulations but can be changed to True to select only glaciated simulations
    min_glacier_depth : float, optional
        Selects simulation with minimum snow height higher than this threshold (in mm)
    max_glacier_depth : float, optional
        Selects simulation with minimum snow height lower than this threshold (in mm)

    Returns
    -------
    pkl : dict
        Dictionary containing all the pickles with keys:
        {df, reanalysis_stats, list_valid_sim, dict_melt_out, stats_melt_out_dic, df_stats, rockfall_values}
        Each pickle is defined by
        df : pandas.core.frame.DataFrame
            A panda dataframe version of the .csv file where the simulation paramaters have been unpacked into readable columns,
            saved to a pickle file
            The 'directory' column is ordered by the simulation index provided in 'path_ground'
        reanalysis_stats : dict
            dictionary of mean quntities over the background and transient periods
        list_valid_sim : list
            list of simulation number for all valid simulations
        dict_melt_out : dict
            Dictionary that assings a melt-out date to each simulation and each year
            Note that if snow does NOT melt at all, we assign the maximal value corresponding to the end of the year 
        stats_melt_out_dic : dict
            Background, transient and full mean of the melt-out date for each simulation
        df_stats : pandas.core.frame.DataFrame
            Large panda dataframe with information about the air, ground, snow, topography, etc. for all simulations
        rockfall_values : dict
            Dictionary with date and topography info for the event

    """

    df = assign_value_df(path_repository, path_ground, path_pickle, site)
    reanalysis_stats = assign_value_reanalysis_stat(forcing_list, path_forcing_list, path_pickle, year_bkg_end, year_trans_end, site)
    list_valid_sim = glacier_filter(site, path_snow, path_pickle, glacier, min_glacier_depth, max_glacier_depth)
    dict_melt_out, stats_melt_out_dic = melt_out_date(consecutive, path_ground, path_snow, path_pickle, year_bkg_end, year_trans_end, site)
    df_stats = assign_value_df_stats(path_ground, path_snow, path_pickle, year_bkg_end, year_trans_end, site)
    rockfall_values = assign_rockfall_values(site, path_pickle, path_ground, path_forcing_list, date_event, topo_event)

    pkl = {'df': df,
           'reanalysis_stats': reanalysis_stats,
           'list_valid_sim': list_valid_sim,
           'dict_melt_out': dict_melt_out,
           'stats_melt_out_dic': stats_melt_out_dic,
           'df_stats': df_stats,
           'rockfall_values': rockfall_values}

    return pkl



"""This module creates seasonal plots with interannual means and variations"""

#pylint: disable=line-too-long
#pylint: disable=trailing-whitespace
#pylint: disable=invalid-name

import pandas as pd
from netCDF4 import num2date #pylint: disable=no-name-in-module
import numpy as np
import matplotlib.pyplot as plt
import mpl_axes_aligner

from SuPerSim.constants import colorcycle, units

def stats_all_years_simulations_to_single_year(time_file, time_series, list_valid_sim, mask_period=None):
    """ Function returns daily mean and several quantiles of a multi-year timeseries over a 1-year period
        e.g. assigns the GST mean over all years and simulations on a given day, e.g. Dec 21st
    
    Parameters
    ----------
    time_file : netCDF4._netCDF4.Variable
        File where the time index of each datapoint is stored (time_ground, not time_air)
    time_series : netCDF4._netCDF4.Variable
        Time series (could be temperature, precipitation, snow depth, etc.)
    list_valid_sim : list
        List of the indices of all valid simulations
    mask_period : numpy.ma.core.MaskedArray, opt
        Period could be 'background' or 'transient' and is a list of booleans, i.e. a mask.
        Can take values 'time_bkg_ground' or 'time_trans_ground' for instance.
        If None, it considers everything.

    Returns
    -------
    quantiles : pandas.core.frame.DataFrame
        Panda dataframe in the shape (5, 366)
        There is a row per quantile in [0.023, 0.16, 0.5, 0.84, 0.977] and a column per timestamp in the year (daily: 366)
    mean_end : pandas.core.series.Series
        One year time series of the mean

    """

    # create a panda dataframe with month, day, hour for each timestamp
    panda_test = pd.DataFrame(num2date(time_file[:], time_file.units), columns=['date'])
    panda_test['month'] = [i.month for i in panda_test['date']]
    panda_test['day'] = [i.day for i in panda_test['date']]
    panda_test['hour'] = [i.hour for i in panda_test['date']]
    panda_test = panda_test.drop(columns=['date'])

    if mask_period is not None:
        panda_test = panda_test.drop(index=np.arange(0,len(mask_period),1)[[not i for i in mask_period]])
 

    mean_test = []
    # for all simulations
    for sim in list_valid_sim:
        if mask_period is None:
            panda_test['timeseries'] = time_series[sim,:,0] if len(time_series.shape) == 3 else time_series[sim,:]
        else:
            panda_test['timeseries'] = time_series[sim,mask_period[:],0] if len(time_series.shape) == 3 else time_series[sim,mask_period[:]]
        # here we group the values of the multi-year timeseries by month, day, and hour and take the mean
        # this produces a single year average timeseries
        mean_test.append([i for i in panda_test.groupby(['month', 'day', 'hour']).mean().reset_index()['timeseries']])
        panda_test = panda_test.drop(columns=['timeseries'])

    mean_test_df = pd.DataFrame(np.array(mean_test))
    # 1-year timeseries of 5 quantiles corresponding to the median plus minus 1 and 2 sigmas
    quantiles = mean_test_df.quantile([0.023, 0.16, 0.5, 0.84, 0.977])
    # 1-year timeseries of the mean
    mean_end = mean_test_df.mean()

    return quantiles, mean_end

def stats_air_all_years_simulations_to_single_year(time_file, time_series, mask_period=None):
    """ Function returns daily mean and several quantiles of a multi-year timeseries over a 1-year period
        for atmospheric quantities (no simulations). The quantiles are over the years, not over simulations.
        e.g. assigns the GST mean over all years and simulations on a given day, e.g. Dec 21st
    
    Parameters
    ----------
    time_file : netCDF4._netCDF4.Variable
        File where the time index of each datapoint is stored (time_ground, not time_air)
    time_series : netCDF4._netCDF4.Variable
        Time series (could be temperature, precipitation, snow depth, etc.)
    mask_period : numpy.ma.core.MaskedArray, opt
        Period could be 'background' or 'transient' and is a list of booleans, i.e. a mask.
        Can take values 'time_bkg_ground' or 'time_trans_ground' for instance.
        If None, it considers everything.

    Returns
    -------
    quantiles : pandas.core.frame.DataFrame
        Panda dataframe in the shape (5, 8784)
        There is a row per quantile in [0.023, 0.16, 0.5, 0.84, 0.977] and a column per timestamp in the year (hourly: 8784)
    mean_end : pandas.core.series.Series
        One year time series of the mean

    """

    # create a panda dataframe with month, day, hour for each timestamp
    panda_test = pd.DataFrame(num2date(time_file[:], time_file.units), columns=['date'])
    panda_test['month'] = [i.month for i in panda_test['date']]
    panda_test['day'] = [i.day for i in panda_test['date']]
    panda_test['hour'] = [i.hour for i in panda_test['date']]
    panda_test = panda_test.drop(columns=['date'])

    if mask_period is None:
        #pylint: disable=unsupported-assignment-operation
        panda_test['timeseries'] = time_series[:,2]
    else:
        panda_test = panda_test.drop(index=np.arange(0,len(mask_period),1)[[not i for i in mask_period]])
        panda_test['timeseries'] = time_series[mask_period[:],2]


    mean_end = pd.DataFrame(np.array([i for i in panda_test.groupby(['month', 'day', 'hour']).mean().reset_index()['timeseries']]))

    list_quant = [0.023, 0.16, 0.5, 0.84, 0.977]

    panda_test = panda_test.groupby(['month', 'day', 'hour']).quantile(list_quant)
    panda_test.index.names = ['month', 'day', 'hour', 'quantiles']
    panda_test = panda_test.reset_index().drop(columns=['month', 'day', 'hour'])

    quantiles = []
    for i in range(len(list_quant)):
        quantiles.append(panda_test.loc[i::len(list_quant)]['timeseries'])

    quantiles = pd.DataFrame(np.array(quantiles))

    return quantiles, mean_end

def data_one_year_quantiles_two_periods(time_file, time_series_list, list_valid_sim_list, list_mask_period=None):
    """ Function returns a single timeseries for 2 periods reduced to a 1-year window with mean and 1 and 2-sigma spread,
    for background and transient piods
    
    Parameters
    ----------
    time_file : netCDF4._netCDF4.Variable
        File where the time index of each datapoint is stored (time_ground, not time_air)
    time_series_list : list of netCDF4._netCDF4.Variable
        List of time series (could be temperature, precipitation, snow depth, etc.)
    list_valid_sim_list : list of list
        List of list of the indices of all valid simulations
    list_mask_period : list
        List of the time masks for each entry, could be 'time_bkg_ground' or 'time_trans_ground' for instance.


    Returns
    -------
    quantiles : list of pandas.core.frame.DataFrame
        list of 2 panda dataframes in the shape (5, number of yearly timestamps), one for each period
        There is a row per quantile in [0.023, 0.16, 0.5, 0.84, 0.977] and a column per timestamp in the year (hourly: ~8784, yearly: ~365)
    mean_end : list of pandas.core.series.Series
        List of 1-year time series of the mean, one for each periods
    """
 
    delta_hours = int((num2date(time_file[1], time_file.units)-num2date(time_file[0], time_file.units)).total_seconds()/3600)

    quantiles = [[] for _ in range(len(time_series_list))]
    mean_end = [[] for _ in range(len(time_series_list))]

    for idx in range(len(time_series_list)):
        if delta_hours == 1:
            quantiles[idx], mean_end[idx] = stats_air_all_years_simulations_to_single_year(time_file, time_series_list[idx], None if list_mask_period is None else list_mask_period[idx])
        else:
            quantiles[idx], mean_end[idx] = stats_all_years_simulations_to_single_year(time_file, time_series_list[idx], list_valid_sim_list[idx], None if list_mask_period is None else list_mask_period[idx])

    return quantiles, mean_end

def plot_sanity_one_year_quantiles_two_periods(quantiles, mean_end, axis_label, list_label):
    """ Function returns a plot of a single timeseries reduced to a 1-year window with mean and 1 and 2-sigma spread,
    for background and transient piods
    
    Parameters
    ----------
    quantiles : list of pandas.core.frame.DataFrame
        list of 2 panda dataframes in the shape (5, number of yearly timestamps), one for each period
        There is a row per quantile in [0.023, 0.16, 0.5, 0.84, 0.977] and a column per timestamp in the year (hourly: ~8784, yearly: ~365)
    mean_end : list of pandas.core.series.Series
        List of 1-year time series of the mean, one for each periods
    axis_label : str
        Label of the y axis, could be 'GST' or 'Snow depth' for instance
    list_label : list
        List of the labels associated to each entry (if both entries are identical, can put a single label)

    Returns
    -------
    fig : Figure
        Plot of 2 timeseries reduced to a 1-year window with mean and 1 and 2-sigma spread.
        Both series have their own y axis if they have different units.
    """

    xdata = range(len(mean_end[0]))

    fig = plt.subplots()

    for idx in range(len(quantiles)):
        plt.plot(xdata, mean_end[idx], color=colorcycle[idx], linewidth=2, label=list_label[idx])
        plt.fill_between(xdata, quantiles[idx].iloc[1], quantiles[idx].iloc[3], alpha = 0.4, color=colorcycle[idx], linewidth=1)
        plt.fill_between(xdata, quantiles[idx].iloc[0], quantiles[idx].iloc[4], alpha = 0.2, color=colorcycle[idx], linewidth=0.5)
        plt.ylabel(axis_label+' ['+units[axis_label]+']')

    if axis_label in ['GST', 'Air temperature']:
        plt.axhline(y=0, color='grey', linestyle='dashed')

    locs, labels = plt.xticks()
    locs = np.linspace(0, len(mean_end[0]), num=13, endpoint=True)
    labels = ['Jan','Feb','Mar','Apr','May','Jun','Jul','Aug','Sep','Oct','Nov','Dec','Jan']
    plt.xticks(locs, labels)

    plt.tight_layout()  # otherwise the right y-label is slightly clipped

    # Show the graph
    plt.legend(loc="upper right")
    plt.show()
    plt.close()

    return fig
    
def plot_sanity_one_year_quantiles_two_periods_from_inputs(time_file, time_series_list, list_valid_sim_list, axis_label, list_label, list_mask_period):
    """ Function returns a plot of a single timeseries reduced to a 1-year window with mean and 1 and 2-sigma spread,
    for background and transient piods
    
    Parameters
    ----------
    time_file : netCDF4._netCDF4.Variable
        File where the time index of each datapoint is stored (time_ground, not time_air)
    time_series_list : list of netCDF4._netCDF4.Variable
        List of time series (could be temperature, precipitation, snow depth, etc.)
    list_valid_sim_list : list of list
        List of list of the indices of all valid simulations
    axis_label : str
        Label of the y axis, could be 'GST' or 'Snow depth' for instance
    list_label : list
        List of the labels associated to each entry (if both entries are identical, can put a single label)
    list_mask_period : list
        List of the time masks for each entry, could be 'time_bkg_ground' or 'time_trans_ground' for instance.

    Returns
    -------
    fig : Figure
        Plot of 2 timeseries reduced to a 1-year window with mean and 1 and 2-sigma spread.
        Both series have their own y axis if they have different units.
    """

    quantiles, mean_end = data_one_year_quantiles_two_periods(time_file, time_series_list, list_valid_sim_list, list_mask_period)
    fig = plot_sanity_one_year_quantiles_two_periods(quantiles, mean_end, axis_label, list_label)

    return fig

def plot_sanity_two_variables_one_year_quantiles(quantiles, mean_end, list_label, list_site=None):
    """ Function returns a plot of 2 timeseries reduced to a 1-year window with mean and 1 and 2-sigma spread.
    
    Parameters
    ----------
    quantiles : list of pandas.core.frame.DataFrame
        list of 2 panda dataframes in the shape (5, number of yearly timestamps), one for each period
        There is a row per quantile in [0.023, 0.16, 0.5, 0.84, 0.977] and a column per timestamp in the year (hourly: ~8784, yearly: ~365)
    mean_end : list of pandas.core.series.Series
        List of 1-year time series of the mean, one for each variable
    list_label : list
        List of the labels associated to each entry (if both entries are identical, can put a single label)
    list_site : list, optional
        List of labels for the site of each entry
        If both labels are identical (i.e. if only 1 label is provided in list_label) then this means we are plotting 
        the same quantity for different sites, in which case we want to label the curves by their site

    Returns
    -------
    fig : Figure
        Plot of 2 timeseries reduced to a 1-year window with mean and 1 and 2-sigma spread.
        Both series have their own y axis if they have different units.
    """
    
    xdata = range(len(mean_end[0]))

    fig, ax1 = plt.subplots()

    idx=0
    ax1.plot(xdata, mean_end[idx], color=colorcycle[idx], linewidth=2, label=list_site[0] if len(list_label) == 1 else list_label[0])
    ax1.fill_between(xdata, quantiles[idx].iloc[1], quantiles[idx].iloc[3], alpha = 0.4, color=colorcycle[idx], linewidth=1)
    ax1.fill_between(xdata, quantiles[idx].iloc[0], quantiles[idx].iloc[4], alpha = 0.2, color=colorcycle[idx], linewidth=0.5)
    ax1.set_ylabel(list_label[0]+' ['+units[list_label[0]]+']')

    idx=1

    if len(list_label) == 1:
        ax1.plot(xdata, mean_end[idx], color=colorcycle[idx], linewidth=2, label=list_site[1])
        ax1.fill_between(xdata, quantiles[idx].iloc[1], quantiles[idx].iloc[3], alpha = 0.4, color=colorcycle[idx], linewidth=1)
        ax1.fill_between(xdata, quantiles[idx].iloc[0], quantiles[idx].iloc[4], alpha = 0.2, color=colorcycle[idx], linewidth=0.5)

    else:
        ax2 = ax1.twinx()
        ax2.plot(xdata, mean_end[idx], color=colorcycle[idx], linewidth=2, label=list_label[1])
        ax2.fill_between(xdata, quantiles[idx].iloc[1], quantiles[idx].iloc[3], alpha = 0.4, color=colorcycle[idx], linewidth=1)
        ax2.fill_between(xdata, quantiles[idx].iloc[0], quantiles[idx].iloc[4], alpha = 0.2, color=colorcycle[idx], linewidth=0.5)
        ax2.set_ylabel(list_label[1]+' ['+units[list_label[1]]+']')

        mpl_axes_aligner.align.yaxes(ax1, 0, ax2, 0, 0.5)

    for i in list_label:
        if i in ['GST', 'Air temperature']:
            plt.axhline(y=0, color='grey', linestyle='dashed')

    locs, labels = plt.xticks()
    locs = np.linspace(0, len(mean_end[0]), num=13, endpoint=True)
    labels = ['Jan','Feb','Mar','Apr','May','Jun','Jul','Aug','Sep','Oct','Nov','Dec','Jan']
    plt.xticks(locs, labels)

    fig.tight_layout()  # otherwise the right y-label is slightly clipped

    # Show the graph
    fig.legend(loc="upper right", bbox_to_anchor=(1,1), bbox_transform=ax1.transAxes)
    plt.show()
    plt.close()

    return fig

def plot_sanity_two_variables_one_year_quantiles_from_inputs(time_file, time_series_list, list_valid_sim_list, list_label, list_site=None):
    """ Function returns a plot of 2 timeseries reduced to a 1-year window with mean and 1 and 2-sigma spread.
    
    Parameters
    ----------
    time_file : netCDF4._netCDF4.Variable
        File where the time index of each datapoint is stored (time_ground, not time_air)
    time_series_list : list of netCDF4._netCDF4.Variable
        List of time series (could be temperature, precipitation, snow depth, etc.)
    list_valid_sim_list : list of list
        List of list of the indices of all valid simulations
    list_label : list
        List of the labels associated to each entry (if both entries are identical, can put a single label)
    list_site : list, optional
        List of labels for the site of each entry
        If both labels are identical (i.e. if only 1 label is provided in list_label) then this means we are plotting 
        the same quantity for different sites, in which case we want to label the curves by their site

    Returns
    -------
    fig : Figure
        Plot of 2 timeseries reduced to a 1-year window with mean and 1 and 2-sigma spread.
        Both series have their own y axis if they have different units.
    """

    quantiles, mean_end = data_one_year_quantiles_two_periods(time_file, time_series_list, list_valid_sim_list)
    fig = plot_sanity_two_variables_one_year_quantiles(quantiles, mean_end, list_label, list_site)

    return fig

def data_two_variables_two_sites_one_year_quantiles(time_file, time_series_list, list_valid_sim_list):
    """ Function returns 2 plots side by side of 2 timeseries each reduced to a 1-year window with mean and 1 and 2-sigma spread.
        Each plot is a plot of two timeseries of the same variable at two different sites
    
    Parameters
    ----------
    time_file : netCDF4._netCDF4.Variable
        File where the time index of each datapoint is stored (time_ground, not time_air)
    time_series_list : list of netCDF4._netCDF4.Variable
        List of time series (could be temperature, precipitation, snow depth, etc.). 1 per plot.
    list_valid_sim_list : list of list
        List of list of the indices of all valid simulations

    Returns
    -------
    quantiles : list of pandas.core.frame.DataFrame
        list of 2*2 panda dataframes in the shape (5, number of yearly timestamps), one for each period
        There is a row per quantile in [0.023, 0.16, 0.5, 0.84, 0.977] and a column per timestamp in the year (hourly: ~8784, yearly: ~365)
        [[df_var0_site0, df_var0_site1], [df_var1_site0, df_var1_site1]]
    mean_end : list of pandas.core.series.Series
        List of 1-year time series of the mean, one for each periods
        [[df_var0_site0, df_var0_site1], [df_var1_site0, df_var1_site1]]
    """

    quantiles = [[] for _ in range(len(time_series_list))]
    mean_end = [[] for _ in range(len(time_series_list))]

    for ts in range(len(time_series_list)):
        quantiles[ts], mean_end[ts] = data_one_year_quantiles_two_periods(time_file, time_series_list[ts], list_valid_sim_list)

    return quantiles, mean_end

def plot_sanity_two_variables_two_sites_one_year_quantiles_side_by_side(quantiles, mean_end, list_label, list_site):
    """ Function returns 2 plots side by side of 2 timeseries each reduced to a 1-year window with mean and 1 and 2-sigma spread.
        Each plot is a plot of two timeseries of the same variable at two different sites
    
    Parameters
    ----------
    quantiles : list of pandas.core.frame.DataFrame
        list of 2*2 panda dataframes in the shape (5, number of yearly timestamps), one for each period
        There is a row per quantile in [0.023, 0.16, 0.5, 0.84, 0.977] and a column per timestamp in the year (hourly: ~8784, yearly: ~365)
        [[df_var0_site0, df_var0_site1], [df_var1_site0, df_var1_site1]]
    mean_end : list of pandas.core.series.Series
        List of 1-year time series of the mean, one for each periods
        [[df_var0_site0, df_var0_site1], [df_var1_site0, df_var1_site1]]
    list_label : list
        List of the labels associated to each plot
    list_site : list
        List of labels for the site of each entry

    Returns
    -------
    fig : Figure
        Plot of 2 timeseries over 2 sites reduced to a 1-year window with mean and 1 and 2-sigma spread.
        1 subplot per variable, both sites compared on the same plot.
    """

    fig, a = plt.subplots(1, 2, figsize=(10, 5))
    for ts,ax in enumerate(a):

        site=0
        xdata = range(len(mean_end[ts][site]))

        ax.plot(xdata, mean_end[ts][site], color=colorcycle[site], linewidth=2, label=list_site[site])
        ax.fill_between(xdata, quantiles[ts][site].iloc[1], quantiles[ts][site].iloc[3], alpha = 0.4, color=colorcycle[site], linewidth=1)
        ax.fill_between(xdata, quantiles[ts][site].iloc[0], quantiles[ts][site].iloc[4], alpha = 0.2, color=colorcycle[site], linewidth=0.5)
        ax.set_ylabel(list_label[ts]+' ['+units[list_label[ts]]+']')

        site=1
        ax.plot(xdata, mean_end[ts][site], color=colorcycle[site], linewidth=2, label=list_site[site])
        ax.fill_between(xdata, quantiles[ts][site].iloc[1], quantiles[ts][site].iloc[3], alpha = 0.4, color=colorcycle[site], linewidth=1)
        ax.fill_between(xdata, quantiles[ts][site].iloc[0], quantiles[ts][site].iloc[4], alpha = 0.2, color=colorcycle[site], linewidth=0.5)

        if list_label[ts] in ['GST', 'Air temperature']:
            ax.axhline(y=0, color='grey', linestyle='dashed')

        locs = np.linspace(0, len(mean_end[ts][site]), num=13, endpoint=True)
        labels = ['Jan','Feb','Mar','Apr','May','Jun','Jul','Aug','Sep','Oct','Nov','Dec','Jan']
        ax.set_xticks(locs)
        ax.set_xticklabels(labels)

        if ts == 1:
            ax.legend(loc="upper right")


    plt.tight_layout()  # otherwise the right y-label is slightly clipped

    # Show the graph
    plt.show()
    plt.close()

    return fig

def plot_sanity_two_variables_two_sites_one_year_quantiles_side_by_side_from_inputs(time_file, time_series_list, list_valid_sim_list, list_label, list_site):
    """ Function returns 2 plots side by side of 2 timeseries each reduced to a 1-year window with mean and 1 and 2-sigma spread.
        Each plot is a plot of two timeseries of the same variable at two different sites
    
    Parameters
    ----------
    time_file : netCDF4._netCDF4.Variable
        File where the time index of each datapoint is stored (time_ground, not time_air)
    time_series_list : list of netCDF4._netCDF4.Variable
        List of time series (could be temperature, precipitation, snow depth, etc.). 1 per plot.
    list_valid_sim_list : list of list
        List of list of the indices of all valid simulations
    list_label : list
        List of the labels associated to each plot
    list_site : list
        List of labels for the site of each entry

    Returns
    -------
    fig : Figure
        Plot of 2 timeseries over 2 sites reduced to a 1-year window with mean and 1 and 2-sigma spread.
        1 subplot per variable, both sites compared on the same plot.
    """

    quantiles, mean_end = data_two_variables_two_sites_one_year_quantiles(time_file, time_series_list, list_valid_sim_list)
    fig = plot_sanity_two_variables_two_sites_one_year_quantiles_side_by_side(quantiles, mean_end, list_label, list_site)

    return fig



"""This module creates running statistics for timerseries"""

#pylint: disable=line-too-long
#pylint: disable=trailing-whitespace
#pylint: disable=invalid-name

import itertools
import pandas as pd
from netCDF4 import num2date #pylint: disable=no-name-in-module
import numpy as np
import matplotlib.pyplot as plt

from SuPerSim.open import open_air_nc, open_ground_nc, open_swe_nc
from SuPerSim.mytime import list_tokens_year
from SuPerSim.weights import assign_weight_sim
from SuPerSim.pickling import load_all_pickles

def mean_all_altitudes(file_to_smooth, site, path_pickle, no_weight=True):
    """ Function returns the mean time series over all altitudes
    
    Parameters
    ----------
    file_to_smooth : netCDF4._netCDF4.Variable
        Atmospheric time series (could be air temperature, precipitation, etc.) that needs to be smoothed
        Needs to have shape (#time, #altitude)
    site : str
        Location of the event, e.g. 'Joffre' or 'Fingerpost'
    path_pickle : str
        String path to the location of the folder where the pickles are saved
    no_weight : bool, optional
        If True, all simulations have the same weight, otherwise the weight is computed as a function of altitude, aspect, and slope

    Returns
    -------
    mean : dict
        average time series over all altitudes
    """

    pkl = load_all_pickles(site, path_pickle)
    df_stats = pkl['df_stats']

    _, pd_weight_long = assign_weight_sim(site, path_pickle, no_weight)

    # list of (altitude, altitude_weight) for all entries of df_stats
    # set -> list of unique entries, then sorted by altitude
    # the result is the weight for each altitude index of the timeseries
    weights = [i[1] for i in sorted(set([(pd_weight_long['altitude'].loc[i], pd_weight_long['altitude_weight'].loc[i]) for i in list(df_stats.index.values)]))]
    
    mean_alt = np.average(file_to_smooth, axis=1, weights=weights)

    return mean_alt

def mean_all_reanalyses(time_files, files_to_smooth, year_bkg_end, year_trans_end):
    """ Function returns the mean time series over a number of reanalysis (has the length of a timeseries)
    
    Parameters
    ----------
    time_files : list of netCDF4._netCDF4.Variable
        List of files where the time index of each datapoint is stored
    files_to_smooth : list of list
        List of time series (could be temperature, precipitation, snow depth, etc.) that needs to be smoothed
        Note that it needs to be in the shape (n,) and not (n,3) for instance, hence the altitude has
        to be pre-selected. E.g. accepts [temp_air_era5[:,0]] but not [temp_air_era5]
    year_bkg_end : int
        Background period is BEFORE the start of the year corresponding to the variable, i.e. all time stamps before Jan 1st year_bkg_end
    year_trans_end : int
        Same for transient period

    Returns
    -------
    mean : dict
        average time series over reanalyses (3 or less)
    """

    list_years = [list_tokens_year(time_files[i], year_bkg_end, year_trans_end)[0] for i in range(len(time_files))]
    len_years = {y: [len(list_years[i][y]) for i in range(len(time_files))] for y in list(list_years[0].keys()) if y<year_trans_end}

    mean = []

    # presumably this means that y goes from 1980 to 2019
    for y in list(len_years.keys()):
        # check if all lists have the same amount of indices in a given year
        order = sorted(len_years[y])
        min_len = order[0]
        max_len = order[-1]
        if min_len == max_len:
            # if True, then we can proceed to computing the mean
            mean += list(np.mean([files_to_smooth[i][list_years[i][y]] for i in range(len(time_files))], axis=0))
        else:
            for j in range(len(order)):
                mean += list(np.mean([files_to_smooth[i][list_years[i][y]][(order[j-1] if j>0 else 0):order[j]]
                                        for i in range(len(time_files)) if len(files_to_smooth[i][list_years[i][y]]) > (order[j-1] if j>0 else 0)], axis=0))

    return mean

def assign_tot_water_prod(path_forcing_list, path_ground, path_swe, path_pickle, year_bkg_end, year_trans_end, site, no_weight=True):
    """ Function returns the total water production at daily intervals in [mm s-1] 
    
    Parameters
    ----------
    path_forcing_list : list of str
        List of paths to the .nc file where the atmospheric forcing data for the given reanalysis is stored
    path_ground : str
        Path to the .nc file where the aggregated ground simulations are stored
    path_swe : str
        Path to the .nc file where the aggregated SWE simulations are stored
    path_pickle : str
        String path to the location of the folder where the pickles are saved
    year_bkg_end : int
        Background period is BEFORE the start of the year corresponding to the variable, i.e. all time stamps before Jan 1st year_bkg_end
    year_trans_end : int
        Same for transient period
    site : str
        Location of the event, e.g. 'Aksaut_North', will be used to label all the pickles
    no_weight : bool, optional
        If True, all simulations have the same weight, otherwise the weight is computed as a function of altitude, aspect, and slope

    Returns
    -------
    tot_water_prod : list
        list of daily total water production
    mean_swe : list
        Weighted mean SWE time series over all simulations
    mean_prec : list
        Mean precipitation time series over all reanalyses
    """

    _, swe = open_swe_nc(path_swe)
    _, time_ground, _ = open_ground_nc(path_ground)
    _, _, _, time_pre_trans_ground = list_tokens_year(time_ground, year_bkg_end, year_trans_end)    

    pd_weight, _ = assign_weight_sim(site, path_pickle, no_weight)

    time_air_all = [open_air_nc(i)[0] for i in path_forcing_list]
    precipitation_all = [open_air_nc(i)[-1] for i in path_forcing_list]

    # here we get the mean precipitation and then water from snow melting 
    mean_swe = list(np.average([swe[i,:] for i in list(pd_weight.index.values)], axis=0, weights=pd_weight.loc[:, 'weight']))
    mean_prec = mean_all_reanalyses(time_air_all,
                                    [mean_all_altitudes(i, site, path_pickle, no_weight) for i in precipitation_all],
                                    year_bkg_end, year_trans_end)

    # convert mean precipitation into a panda dataframe to facilitate grouping by 24.
    # This way the data is reported daily (as it is for the swe) rather than hourly
    pd_prec_day = ((pd.DataFrame(mean_prec).rolling(24).mean()).iloc[23::24, :]).reset_index(drop=True)

    # convert swe into panda df to make the diff() operation easier. Swe is a total and
    # we get the instantaneous flux by taking the difference between consecutive datapoints.
    # The total precipitation between 2 datapoints is over 1 day = 864000 hence we divide by 86400.
    # Finally, swe is in [kg m-2] which is equivalent to [mm] so we're good.

    # Number of days between the start of the ground and air timeseries
    num_day = int((num2date(time_ground[0], time_ground.units) - num2date(time_air_all[0][0], time_air_all[0].units)).total_seconds()/86400)
    # we add num_days zeros at the beginning to make sure the dfs have the same dimension and span Jan 1, 1980 to Dec 31, 2019
    # We make sure to multiply by (-1) and to divide by 86400
    pd_diff_swe = pd.concat([pd.DataFrame([0]*num_day), (pd.DataFrame(mean_swe[:len(time_ground[time_pre_trans_ground][:])]).diff()).fillna(0)/86400*(-1)], ignore_index=True)

    if len(pd_prec_day)!=len(pd_diff_swe):
        raise ValueError('VictorCustomError: The lengths of the precipitation and swe time series are different!')

    # The total water production is returned in the list type as the sum of precipitation and swe
    tot_water_prod = list(pd_prec_day.add(pd_diff_swe.reset_index(drop=True), fill_value=0)[0])

    return tot_water_prod, mean_swe, mean_prec

def running_mean_std(time_file, file_to_smooth, window, fill_before=False):
    """ Function returns a running average (and standard deviation) of the time series over a given time window
        adpated from Welford's online algorithm, computing everything in a single-pass 
    
    Parameters
    ----------
    time_file : netCDF4._netCDF4.Variable
        File where the time index of each datapoint is stored
    file_to_smooth : list
        Time series (could be temperature, precipitation, snow depth, etc.) that needs to be smoothed
        Note that it needs to be in the shape (n,) and not (n,3) for instance, hence the altitude has
        to be pre-selected. E.g. accepts temp_air_era5[:,0] but not temp_air_era5
    window : str
        Time window of datapoints to average over.
        Can take the following arguments: 'day', 'week', 'year', 'climate'
    fill_before : bool, optional
        Option to fill the first empty values of the smoothed data with the first running average value if True

    Returns
    -------
    running_mean : dict
        Smoothed time series
    running_std : dict
        Running standard deviation

    """

    # number of hours between consecutive measurements: 1 for air, 24 for ground
    cons_hours = int((num2date(time_file[1], time_file.units) - num2date(time_file[0], time_file.units)).total_seconds()/3600)
    window_size = int({'day': 24, 'week': 24*7, 'month': 24*30, 'year':24*365, 'climate':24*35*20}[window]/cons_hours)

    
    # range of the timestamp whether the user chose a given year of the whole study period
    index_range = list(range(len(file_to_smooth)))

    # initialize the running mean, the first value is at (window_size-1+index_range[0])
    # the initial window extends from index_range[0] to window_size-1+index_range[0] and has (window_size) elements
    init = np.sum(file_to_smooth[index_range[0]:window_size+index_range[0]])/window_size
    running_mean = {}
    running_std = {}

    # fill the first values if fill_before==True
    if fill_before:
        for i in range(index_range[0], window_size-1+index_range[0]):
            running_mean[i] = init
            # running_std[i] = 0
    
    running_mean[window_size-1+index_range[0]] = init
    # running_std[window_size-1+index_range[0]] = np.std(file_to_smooth[index_range[0]:window_size+index_range[0]])

    # the following values are evaluated recursively in a single-pass (more efficient)
    # I have derived the running version of the Welford's online algorithm and tested it.
    for i in range(window_size+index_range[0], index_range[-1]+1):
        running_mean[i] = running_mean[i-1] + (file_to_smooth[i] - file_to_smooth[i-window_size])/window_size
        # running_std[i] = np.sqrt( (window_size * running_std[i-1]**2
        #                            + (file_to_smooth[i] - file_to_smooth[i-window_size])**2/window_size
        #                            + (file_to_smooth[i] - running_mean[i])**2
        #                            - (file_to_smooth[i-window_size] - running_mean[i])**2) / window_size )

    return running_mean, running_std

def aggregating_distance_temp(time_file, file_to_smooth, window, year_bkg_end, year_trans_end, year=0, fill_before=False):
    """ Function returns the distance to the mean in units of standard deviation for a given date yy-mm-dd
        relative to the ensemble of same date for previous years: {year-mm-dd for year<=yy}
        Should only be applied to a previously 'smoothed' function through a running average
    
    Parameters
    ----------
    time_file : netCDF4._netCDF4.Variable
        File where the time index of each datapoint is stored        
    file_to_smooth : list
        Time series (could be temperature, precipitation, snow depth, etc.) that needs to be smoothed
        Note that it needs to be in the shape (n,) and not (n,3) for instance, hence the altitude has
        to be pre-selected. E.g. accepts temp_air_era5[:,0] but not temp_air_era5
    window : str
        Time window of datapoints to average over.
        Can take the following arguments: 'day', 'week', 'year', 'climate'
    year_bkg_end : int, optional
        Background period is BEFORE the start of the year corresponding to the variable, i.e. all time stamps before Jan 1st year_bkg_end
    year_trans_end : int, optional
        Same for transient period
    year : int, optional
        One can choose to display a specific year or the whole set by choosing an integer not in the study sample, e.g. 0.
    fill_before : bool, optional
        Option to fill the first empty values of the smoothed data with the first running average value if True

    Returns
    -------
    distance : dict
        Distance from mean in units of standard deviation

    """

    # here temp_smoothed is the smoothed temperature data series in the form of a dictionary 
    # keys are timestamps (end of window) and values are running average of temperatures over window
    temp_smoothed = running_mean_std(time_file, file_to_smooth, window, fill_before)[0]

    list_dates = list_tokens_year(time_file, year_bkg_end, year_trans_end)[0]

    num_min = np.min([len(list_dates[year]) for year in list_dates.keys() if year < year_trans_end])
    min_temp_index = list(temp_smoothed.keys())[0]

    # range of the timestamp whether the user chose a given year of the whole study period
    # index_range = list(range(len(file_to_smooth)))

    aggregate_mean = {}
    aggreagte_std = {}
    distance = {}

    start_year = num2date(time_file[0], time_file.units).year

    for year in [year_bkg_end]:
        aggregate_mean[year] = {day: np.mean([temp_smoothed[list_dates[y][day]] for y in range(start_year,year+1) if list_dates[y][day]>=min_temp_index]) for day in range(num_min)}
        aggreagte_std[year] = {day: np.std([temp_smoothed[list_dates[y][day]] for y in range(start_year,year+1) if list_dates[y][day]>=min_temp_index]) for day in range(num_min)}
        distance[year] = {day: (temp_smoothed[list_dates[year][day]]-aggregate_mean[year][day])/aggreagte_std[year][day] for day in range(num_min)}

    for year in range(year_bkg_end+1, year_trans_end):
        aggregate_mean[year] = {day: aggregate_mean[year-1][day]
                                     + (temp_smoothed[list_dates[year][day]] - aggregate_mean[year-1][day]) / (year-start_year+1)
                                     for day in range(num_min)}
        aggreagte_std[year] = {day: np.sqrt( ( (year-start_year) * aggreagte_std[year-1][day]**2
                                              + (temp_smoothed[list_dates[year][day]] - aggregate_mean[year][day])
                                              *(temp_smoothed[list_dates[year][day]] - aggregate_mean[year-1][day]) )
                                              / (year-start_year+1) )
                                    for day in range(num_min)}
        distance[year] = {day: (temp_smoothed[list_dates[year][day]]-aggregate_mean[year][day])/aggreagte_std[year][day] for day in range(num_min)}


    return distance

def aggregating_distance_temp_all(yaxes, xdata, ydata, window, site, path_pickle, year_bkg_end, year_trans_end, year, fill_before=False):
    """ Returns the distance to the mean in units of standard deviation for a specific year or for the whole length
        and for a list of windows
    
    Parameters
    ----------
    yaxes : list of str
        List of names of quantities to plot, e.g. ['Air temperature', 'Water production', 'Ground temperature', 'Depth of thaw']
        Will be used for labels
    xdata : list of netCDF4._netCDF4.Variable
        List of files where the time index of each datapoint is stored, e.g. [time_air_era5, time_air_era5, time_ground, time_ground]
        len(xdata) should be equal to len(yaxes) 
    ydata : list of list
        List of time series (could be temperature, precipitation, snow depth, etc.) that needs to be smoothed
        Note that each individual time series needs to be in the shape (n,) and not (n,3) for instance, hence the altitude has
        to be pre-selected. E.g. accepts temp_air_era5[:,0] but not temp_air_era5
        example: [mean_air_temp, mean_air_temp, temp_ground_mean, temp_ground_mean]
        len(ydata) should be equal to len(yaxes)
    window : list of str
        Time window of datapoints to average over.
        Can take the following arguments: 'day', 'week', 'year', 'climate'
    site : str
        Location of the event, e.g. 'Joffre' or 'Fingerpost'
    path_pickle : str
        String path to the location of the folder where the pickles are saved
    year_bkg_end : int
        Background period is BEFORE the start of the year corresponding to the variable, i.e. all time stamps before Jan 1st year_bkg_end
    year_trans_end : int
        Same for transient period
    year : int
        One can choose to display a specific year or the whole set by choosing an integer not in the study sample, e.g. 0.
    fill_before : bool, optional
        Option to fill the first empty values of the smoothed data with the first running average value if True

    Returns
    -------
    dict_distances : dict
        Dictionary containing the normalized distances, in the form e.g.
        {'Air temperature': {'week': [...], 'month': [...], ...}, 'Water production': {'week': [...], 'month': [...], ...}, ...}
    rockfall_time_index : dict
        Dictionary listing the time index of the rockfall (if the rockfall happened within the plotting wiwndow), in the form e.g.
        {'Air temperature': 8736, 'Water production': 364, 'Ground temperature': 364}
    """

    pkl = load_all_pickles(site, path_pickle)
    rockfall_values = pkl['rockfall_values']

    dict_distances = {var: {win: [] for win in window} for var in yaxes}
    dict_index_range = {var: [] for var in yaxes}
    dict_list_dates = {var: [] for var in yaxes}
    rockfall_time_index = {var: [] for var in yaxes}

    for idx, win in itertools.product(range(len(yaxes)), window):
        distance = aggregating_distance_temp(xdata[idx], ydata[idx], win, year_bkg_end, year_trans_end, year, fill_before)
        list_dates = list_tokens_year(xdata[idx], year_bkg_end, year_trans_end)[0]
        min_range = list_dates[np.min(list(distance.keys()))][0]
        max_range = list_dates[np.max(list(distance.keys()))][-1]
        #pylint: disable=consider-iterating-dictionary
        index_range = list_dates[year] if year in distance.keys() else list(range(min_range, max_range+1))

        # either we get 1 year of data or the whole dataset (flatten the list)
        #pylint: disable=consider-iterating-dictionary
        dict_distances[yaxes[idx]][win] = list(distance[year].values()) if year in distance.keys() else [i for j in distance.values() for i in j.values()]
        dict_index_range[yaxes[idx]] = list(range(len(ydata[idx])))[list_dates[np.min(list(distance.keys()))][0]:]
        dict_list_dates[yaxes[idx]] = list_dates

        if rockfall_values['exact_date']:
            for time_idx in rockfall_values['time_index']:
                if time_idx in index_range:
                    if int((num2date(xdata[idx][time_idx], xdata[idx].units)-rockfall_values['datetime']).total_seconds()) == 0:
                        rockfall_time_index[yaxes[idx]] = time_idx - index_range[0]

    return dict_distances, rockfall_time_index

def plot_aggregating_distance_temp_all(dict_distances, rockfall_time_index, year_bkg_end, year_trans_end, year):
    """ Plots the distance to the mean in units of standard deviation for a specific year or for the whole length
        Vertical subplots for different variables
        Plots from user-given dictionaries
    
    Parameters
    ----------
    dict_distances : dict
        Dictionary containing the normalized distances, in the form e.g.
        {'Air temperature': {'week': [...], 'month': [...], ...}, 'Water production': {'week': [...], 'month': [...], ...}, ...}
    rockfall_time_index : dict
        Dictionary listing the time index of the rockfall (if the rockfall happened within the plotting wiwndow), in the form e.g.
        {'Air temperature': 8736, 'Water production': 364, 'Ground temperature': 364}
    year_bkg_end : int
        Background period is BEFORE the start of the year corresponding to the variable, i.e. all time stamps before Jan 1st year_bkg_end
    year_trans_end : int
        Same for transient period
    year : int
        One can choose to display a specific year or the whole set by choosing an integer not in the study sample, e.g. 0.

    Returns
    -------
    fig : Figure
        normalized distance plot containing subplots sharing the x and y axis
    """

    yaxes = list(dict_distances.keys())
    window = list(list(dict_distances.values())[0].keys())

    num_rows = len(yaxes)
    num_cols = len(window)

    fig, a = plt.subplots(num_rows, num_cols, figsize=(8, 2*num_rows), sharey='row')
    for idx,ax in enumerate(a):
        # idx is the row index, hence it labels yaxes and ydata
        # ax labels the xdata and window
        if num_cols == 1:
            ax.plot(dict_distances[yaxes[idx]][window[0]], label='Deviation')
            ax.fill_between(range(len(dict_distances[yaxes[idx]][window[0]])), -2, 2, alpha = 0.2, color = 'blue')
            ax.axhline(y = 0, color = 'black', linestyle='--', linewidth=1)
            if not isinstance(rockfall_time_index[yaxes[idx]], list):
                ax.axvline(x = rockfall_time_index[yaxes[idx]], color = 'r', linestyle='--', label = 'Landslide')
        else:
            for i in range(num_cols):
                ax[i].plot(dict_distances[yaxes[idx]][window[i]], label='Deviation')
                ax[i].fill_between(range(len(dict_distances[yaxes[idx]][window[i]])), -2, 2, alpha = 0.2, color = 'blue')
                ax[i].axhline(y = 0, color = 'black', linestyle='--', linewidth=1)
                if not isinstance(rockfall_time_index[yaxes[idx]], list):
                    ax[i].axvline(x = rockfall_time_index[yaxes[idx]], color = 'r', linestyle='--', label = 'Landslide')

        if year in np.arange(year_bkg_end, year_trans_end):
            locs = np.linspace(0, len(dict_distances[yaxes[idx]][window[0]]), num=12, endpoint=False)
            if num_cols == 1:
                labels = ['Jan','Feb','Mar','Apr','May','Jun','Jul','Aug','Sep','Oct','Nov','Dec']
            else:
                labels = ['Jan','','Mar','','May','','Jul','','Sep','','Nov','']
        else:
            locs = np.linspace(0, len(dict_distances[yaxes[idx]][window[0]]), num= year_trans_end - year_bkg_end + 1, endpoint=True)
            dloc = int(len(locs)/10*num_cols)
            locs = locs[::dloc]
            labels = list(range(year_bkg_end,year_trans_end+1,1))
            labels = labels[::dloc]

        if idx < num_rows-1:
            labels_end = ['' for _ in labels]
        else:
            labels_end = labels
        if num_cols == 1:
            ax.set_xticks(locs, labels)
            ax.set_ylabel(yaxes[idx])
            ax.grid(axis = 'y')
        else: 
            for i in range(num_cols):
                ax[i].set_xticks(locs, labels_end)
                ax[i].grid(axis = 'y')
                if idx == 0:
                    ax[i].set_title(window[i].capitalize())
            ax[0].set_ylabel(yaxes[idx])
            ax[1].yaxis.set_tick_params(labelleft=False)
        
    fig.supylabel(r'Normalized deviation $d_{norm}$ [$\sigma$]')
    plt.legend(loc='lower right')
    plt.tight_layout()
    plt.show()
    plt.close()

    return fig

def plot_aggregating_distance_temp_all_from_input(yaxes, xdata, ydata, window, site, path_pickle, year_bkg_end, year_trans_end, year, fill_before=False):
    """ Plots the distance to the mean in units of standard deviation for a specific year or for the whole length
        Vertical subplots for different variables
        Plots directly from intended inputs
    
    Parameters
    ----------
    yaxes : list of str
        List of names of quantities to plot, e.g. ['Air temperature', 'Water production', 'Ground temperature', 'Depth of thaw']
        Will be used for labels
    xdata : list of netCDF4._netCDF4.Variable
        List of files where the time index of each datapoint is stored, e.g. [time_air_era5, time_air_era5, time_ground, time_ground]
        len(xdata) should be equal to len(yaxes) 
    ydata : list of list
        List of time series (could be temperature, precipitation, snow depth, etc.) that needs to be smoothed
        Note that each individual time series needs to be in the shape (n,) and not (n,3) for instance, hence the altitude has
        to be pre-selected. E.g. accepts temp_air_era5[:,0] but not temp_air_era5
        example: [mean_air_temp, mean_air_temp, temp_ground_mean, temp_ground_mean]
        len(ydata) should be equal to len(yaxes)
    window : list of str
        Time window of datapoints to average over.
        Can take the following arguments: 'day', 'week', 'year', 'climate'
    site : str
        Location of the event, e.g. 'Joffre' or 'Fingerpost'
    path_pickle : str
        String path to the location of the folder where the pickles are saved
    year_bkg_end : int
        Background period is BEFORE the start of the year corresponding to the variable, i.e. all time stamps before Jan 1st year_bkg_end
    year_trans_end : int
        Same for transient period
    year : int
        One can choose to display a specific year or the whole set by choosing an integer not in the study sample, e.g. 0.
    fill_before : bool, optional
        Option to fill the first empty values of the smoothed data with the first running average value if True

    Returns
    -------
    fig : Figure
        normalized distance plot containing subplots sharing the x and y axis
    """

    dict_distances, rockfall_time_index = aggregating_distance_temp_all(yaxes, xdata, ydata, window, site, path_pickle, year_bkg_end, year_trans_end, year, fill_before)
    fig = plot_aggregating_distance_temp_all(dict_distances, rockfall_time_index, year_bkg_end, year_trans_end, year)

    return fig
